from pathlib import Path

import scrapli

from .helper import (
    arista_eos_clean_response,
    cisco_iosxe_clean_response,
    cisco_iosxr_clean_response,
    cisco_nxos_clean_response,
    juniper_junos_clean_response,
)

TEST_DATA_PATH = f"{Path(scrapli.__file__).parents[1]}/tests/functional/test_data"


TEST_CASES = {
    "cisco_iosxe": {
        "get_prompt": {
            "exec": "csr1000v>",
            "privilege_exec": "csr1000v#",
            "configuration": "csr1000v(config)#",
        },
        "send_command_short": {
            "command": "show run | i hostname",
            "expected_no_strip": "hostname csr1000v\ncsr1000v#",
            "expected_strip": "hostname csr1000v",
        },
        "send_command_long": {
            "command": "show run",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_iosxe/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_iosxe/send_command_long_strip"
            ).read(),
        },
        "send_commands_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_iosxe/send_commands",
            "expected_no_strip": ["hostname csr1000v\ncsr1000v#", "hostname csr1000v\ncsr1000v#"],
            "expected_strip": ["hostname csr1000v", "hostname csr1000v"],
        },
        "send_commands_error": {
            "commands": ["show version", "show tacocat", "show version"],
        },
        "send_interactive_normal_response": {
            "command": [("clear logg", "Clear logging buffer [confirm]"), ("", "csr1000v#")],
            "expected": "clear logg\nClear logging buffer [confirm]\n\ncsr1000v#",
        },
        "send_interactive_hidden_response": None,
        "send_config": {
            "configs": "interface loopback123\ndescription scrapli was here",
            "expected_no_strip": "csr1000v(config-if)#\ncsr1000v(config-if)#",
            "expected_strip": "\n",
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "Building configuration...\n\nCurrent configuration : CONFIG_BYTES"
            "\n!\ninterface Loopback123\n description scrapli was here\n no ip"
            " address\nend\n\ncsr1000v#",
            "verification_expected_strip": "Building configuration...\n\nCurrent configuration : CONFIG_BYTES"
            "\n!\ninterface Loopback123\n description scrapli was here\n no ip "
            "address\nend",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs": {
            "configs": ["interface loopback123", "description scrapli was here"],
            "expected_no_strip": ["csr1000v(config-if)#", "csr1000v(config-if)#"],
            "expected_strip": ["", ""],
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "Building configuration...\n\nCurrent configuration : CONFIG_BYTES"
            "\n!\ninterface Loopback123\n description scrapli was here\n no ip"
            " address\nend\n\ncsr1000v#",
            "verification_expected_strip": "Building configuration...\n\nCurrent configuration : CONFIG_BYTES"
            "\n!\ninterface Loopback123\n description scrapli was here\n no ip "
            "address\nend",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_iosxe/send_configs",
            "expected_no_strip": ["csr1000v(config-if)#", "csr1000v(config-if)#"],
            "expected_strip": ["", ""],
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_error": {
            "configs": ["interface loopback123", "show tacocat", "description tacocat was here"],
            "teardown_configs": "no interface loopback123",
        },
        "sanitize_response": cisco_iosxe_clean_response,
    },
    "cisco_nxos": {
        "get_prompt": {
            "exec": None,
            "privilege_exec": "switch#",
            "configuration": "switch(config)#",
        },
        "send_command_short": {
            "command": "show run | i scp-server",
            "expected_no_strip": "feature scp-server\nswitch#",
            "expected_strip": "feature scp-server",
        },
        "send_command_long": {
            "command": "show run",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_nxos/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_nxos/send_command_long_strip"
            ).read(),
        },
        "send_commands_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_nxos/send_commands",
            "expected_no_strip": ["feature scp-server\nswitch#", "feature scp-server\nswitch#"],
            "expected_strip": ["feature scp-server", "feature scp-server"],
        },
        "send_commands_error": {
            "commands": ["show version", "show tacocat", "show version"],
        },
        "send_interactive_normal_response": {
            "command": [
                ("delete bootflash:virtual-instance.conf", "(yes/no/abort)   [y]"),
                ("n", "switch#"),
            ],
            "expected": 'delete bootflash:virtual-instance.conf\nDo you want to delete "/virtual-instance.conf" ? (yes/no/abort)   [y] n\nswitch#',
        },
        "send_interactive_hidden_response": None,
        "send_config": {
            "configs": "interface loopback123\ndescription scrapli was here",
            "expected_no_strip": "switch(config-if)#\nswitch(config-if)#",
            "expected_strip": "\n",
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "!Command: show running-config interface loopback123\n!Running "
            "configuration last done at: TIME_STAMP_REPLACED\n!Time: "
            "TIME_STAMP_REPLACED\n\nversion 9.2(4) Bios:version\n\ninterface "
            "loopback123\n  description scrapli was here\n\nswitch#",
            "verification_expected_strip": "!Command: show running-config interface loopback123\n!Running "
            "configuration last done at: TIME_STAMP_REPLACED\n!Time: "
            "TIME_STAMP_REPLACED\n\nversion 9.2(4) Bios:version\n\ninterface "
            "loopback123\n  description scrapli was here",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs": {
            "configs": ["interface loopback123", "description scrapli was here"],
            "expected_no_strip": ["switch(config-if)#", "switch(config-if)#"],
            "expected_strip": ["", ""],
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "!Command: show running-config interface loopback123\n!Running "
            "configuration last done at: TIME_STAMP_REPLACED\n!Time: "
            "TIME_STAMP_REPLACED\n\nversion 9.2(4) Bios:version\n\ninterface "
            "loopback123\n  description scrapli was here\n\nswitch#",
            "verification_expected_strip": "!Command: show running-config interface loopback123\n!Running "
            "configuration last done at: TIME_STAMP_REPLACED\n!Time: "
            "TIME_STAMP_REPLACED\n\nversion 9.2(4) Bios:version\n\ninterface "
            "loopback123\n  description scrapli was here",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_nxos/send_configs",
            "expected_no_strip": ["switch(config-if)#", "switch(config-if)#"],
            "expected_strip": ["", ""],
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_error": {
            "configs": ["interface loopback123", "show tacocat", "description tacocat was here"],
            "teardown_configs": "no interface loopback123",
        },
        "sanitize_response": cisco_nxos_clean_response,
    },
    "cisco_iosxr": {
        "get_prompt": {
            "exec": None,
            "privilege_exec": "RP/0/RP0/CPU0:ios#",
            "configuration": "RP/0/RP0/CPU0:ios(config)#",
        },
        "send_command_short": {
            "command": "show run | i MgmtEth0",
            "expected_no_strip": "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0\nRP/0/RP0/CPU0:ios#",
            "expected_strip": "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0",
        },
        "send_command_long": {
            "command": "show run",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_iosxr/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/cisco_iosxr/send_command_long_strip"
            ).read(),
        },
        "send_commands_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_iosxr/send_commands",
            "expected_no_strip": [
                "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0\nRP/0/RP0/CPU0:ios#",
                "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0\nRP/0/RP0/CPU0:ios#",
            ],
            "expected_strip": [
                "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0",
                "TIME_STAMP_REPLACED\nBuilding configuration...\ninterface MgmtEth0/RP0/CPU0/0",
            ],
        },
        "send_commands_error": {
            "commands": ["show version", "show tacocat", "show version"],
        },
        "send_interactive_normal_response": None,
        "send_interactive_hidden_response": None,
        "send_config": {
            "configs": "interface loopback123\ndescription scrapli was here\ncommit",
            "expected_no_strip": "RP/0/RP0/CPU0:ios(config-if)#\nRP/0/RP0/CPU0:ios(config-if)#\nTIME_STAMP_REPLACED\nRP/0/RP0/CPU0:ios(config-if)#",
            "expected_strip": "\n\nTIME_STAMP_REPLACED",  # we get the timestamp of the commit in this output
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "TIME_STAMP_REPLACED\ninterface Loopback123\n description scrapli was here\n!\n\nRP/0/RP0/CPU0:ios#",
            "verification_expected_strip": "TIME_STAMP_REPLACED\ninterface Loopback123\n description scrapli was here\n!",
            "teardown_configs": ["no interface loopback123", "commit"],
        },
        "send_configs": {
            "configs": ["interface loopback123", "description scrapli was here", "commit"],
            "expected_no_strip": ["RP/0/RP0/CPU0:ios(config-if)#", "RP/0/RP0/CPU0:ios(config-if)#"],
            "expected_strip": ["", ""],
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "TIME_STAMP_REPLACED\ninterface Loopback123\n description scrapli was here\n!\n\nRP/0/RP0/CPU0:ios#",
            "verification_expected_strip": "TIME_STAMP_REPLACED\ninterface Loopback123\n description scrapli was here\n!",
            "teardown_configs": ["no interface loopback123", "commit"],
        },
        "send_configs_from_file": {
            "file": f"{TEST_DATA_PATH}/source/cisco_iosxr/send_configs",
            "expected_no_strip": ["RP/0/RP0/CPU0:ios(config-if)#", "RP/0/RP0/CPU0:ios(config-if)#"],
            "expected_strip": ["", ""],
            "teardown_configs": ["no interface loopback123", "commit"],
        },
        "send_configs_error": {
            "configs": ["interface loopback123", "show tacocat", "description tacocat was here"],
            "teardown_configs": ["no interface loopback123", "commit"],
        },
        "sanitize_response": cisco_iosxr_clean_response,
    },
    "arista_eos": {
        "get_prompt": {
            "exec": "localhost>",
            "privilege_exec": "localhost#",
            "configuration": "localhost(config)#",
        },
        "send_command_short": {
            "command": "show run | i ZTP",
            "expected_no_strip": "logging level ZTP informational\nlocalhost#",
            "expected_strip": "logging level ZTP informational",
        },
        "send_command_long": {
            "command": "show run",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/arista_eos/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/arista_eos/send_command_long_strip"
            ).read(),
        },
        "send_commands_from_file": {
            "file": f"{TEST_DATA_PATH}/source/arista_eos/send_commands",
            "expected_no_strip": [
                "logging level ZTP informational\nlocalhost#",
                "logging level ZTP informational\nlocalhost#",
            ],
            "expected_strip": [
                "logging level ZTP informational",
                "logging level ZTP informational",
            ],
        },
        "send_commands_error": {
            "commands": ["show version", "show tacocat", "show version"],
        },
        "send_interactive_normal_response": None,
        "send_interactive_hidden_response": None,
        "send_config": {
            "configs": "interface loopback123\ndescription scrapli was here",
            "expected_no_strip": "localhost(config-if-Lo123)#\nlocalhost(config-if-Lo123)#",
            "expected_strip": "\n",
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "interface Loopback123\n   description scrapli was here\nlocalhost#",
            "verification_expected_strip": "interface Loopback123\n   description scrapli was here",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs": {
            "configs": ["interface loopback123", "description scrapli was here"],
            "expected_no_strip": ["localhost(config-if-Lo123)#", "localhost(config-if-Lo123)#"],
            "expected_strip": ["", ""],
            "verification": "show run int loopback123",
            "verification_expected_no_strip": "interface Loopback123\n   description scrapli was here\nlocalhost#",
            "verification_expected_strip": "interface Loopback123\n   description scrapli was here",
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_from_file": {
            "file": f"{TEST_DATA_PATH}/source/arista_eos/send_configs",
            "expected_no_strip": ["localhost(config-if-Lo123)#", "localhost(config-if-Lo123)#"],
            "expected_strip": ["", ""],
            "teardown_configs": "no interface loopback123",
        },
        "send_configs_error": {
            "configs": ["interface loopback123", "show tacocat", "description tacocat was here"],
            "teardown_configs": "no interface loopback123",
        },
        "sanitize_response": arista_eos_clean_response,
    },
    "juniper_junos": {
        "get_prompt": {
            "exec": "vrnetlab>",
            "privilege_exec": None,
            "configuration": "vrnetlab#",
        },
        "send_command_short": {
            "command": "show configuration | match 10.0.0.15",
            "expected_no_strip": "                address 10.0.0.15/24;\n\nvrnetlab>",
            "expected_strip": "                address 10.0.0.15/24;",
        },
        "send_command_long": {
            "command": "show configuration",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/juniper_junos/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/juniper_junos/send_command_long_strip"
            ).read(),
        },
        "send_commands_from_file": {
            "file": f"{TEST_DATA_PATH}/source/juniper_junos/send_commands",
            "expected_no_strip": [
                "                address 10.0.0.15/24;\n\nvrnetlab>",
                "                address 10.0.0.15/24;\n\nvrnetlab>",
            ],
            "expected_strip": [
                "                address 10.0.0.15/24;",
                "                address 10.0.0.15/24;",
            ],
        },
        "send_commands_error": {
            "commands": ["show version", "show tacocat", "show version"],
        },
        "send_interactive_normal_response": None,
        "send_interactive_hidden_response": None,
        "send_config": {
            "configs": 'set interfaces fxp0 unit 0 description "scrapli was here"\ncommit',
            "expected_no_strip": "[edit]\nvrnetlab#\ncommit complete\n\n[edit]\nvrnetlab#",
            "expected_strip": "[edit]\ncommit complete\n\n[edit]",
            "verification": "show configuration interfaces fxp0",
            "verification_expected_no_strip": 'unit 0 {\n    description "scrapli was here";\n    family inet {\n        address 10.0.0.15/24;\n    }\n}\n\nvrnetlab>',
            "verification_expected_strip": 'unit 0 {\n    description "scrapli was here";\n    family inet {\n        address 10.0.0.15/24;\n    }\n}',
            "teardown_configs": ["delete interfaces fxp0 unit 0 description", "commit"],
        },
        "send_configs": {
            "configs": ['set interfaces fxp0 unit 0 description "scrapli was here"', "commit"],
            "expected_no_strip": ["[edit]\nvrnetlab#", "commit complete\n\n[edit]\nvrnetlab#"],
            "expected_strip": ["[edit]", "commit complete\n\n[edit]"],
            "verification": "show configuration interfaces fxp0",
            "verification_expected_no_strip": 'unit 0 {\n    description "scrapli was here";\n    family inet {\n        address 10.0.0.15/24;\n    }\n}\n\nvrnetlab>',
            "verification_expected_strip": 'unit 0 {\n    description "scrapli was here";\n    family inet {\n        address 10.0.0.15/24;\n    }\n}',
            "teardown_configs": ["delete interfaces fxp0 unit 0 description", "commit"],
        },
        "send_configs_from_file": {
            "file": f"{TEST_DATA_PATH}/source/juniper_junos/send_configs",
            "expected_no_strip": ["[edit]\nvrnetlab#", "commit complete\n\n[edit]\nvrnetlab#"],
            "expected_strip": ["[edit]", "commit complete\n\n[edit]"],
            "teardown_configs": ["delete interfaces fxp0 unit 0 description", "commit"],
        },
        "send_configs_error": {
            "configs": [
                "set interfaces fxp0 description tacocat",
                "show tacocat",
                "set interfaces fxp0 description tacocat",
            ],
            "teardown_configs": ["delete interfaces fxp0 description", "commit"],
        },
        "sanitize_response": juniper_junos_clean_response,
    },
    "linux": {
        "get_prompt": "linux:~#",
        "send_command_short": {
            "command": "cat /etc/hostname",
            "expected_no_strip": "linux\nlinux:~#",
            "expected_strip": "linux",
        },
        "send_command_long": {
            "command": "cat /etc/init.d/sshd",
            "expected_no_strip": open(
                f"{TEST_DATA_PATH}/expected/linux/send_command_long_no_strip"
            ).read(),
            "expected_strip": open(
                f"{TEST_DATA_PATH}/expected/linux/send_command_long_strip"
            ).read(),
        },
    },
}



import pytest

from scrapli.transport import AsyncTransport

from ...test_data.unit_test_cases import TEST_CASES


def test_init(async_cisco_iosxe_conn):
    """Test that all arguments get properly passed from driver/transport to channel on init"""
    assert isinstance(async_cisco_iosxe_conn.channel.transport, AsyncTransport)
    assert (
        async_cisco_iosxe_conn.channel.comms_prompt_pattern
        == async_cisco_iosxe_conn._initialization_args["comms_prompt_pattern"]
    )
    assert (
        async_cisco_iosxe_conn.channel.comms_return_char
        == async_cisco_iosxe_conn._initialization_args["comms_return_char"]
    )
    assert (
        async_cisco_iosxe_conn.channel.comms_ansi
        == async_cisco_iosxe_conn._initialization_args["comms_ansi"]
    )
    # auto expand is not set in initialization args at this point (may 2020)
    assert (
        async_cisco_iosxe_conn.channel.comms_auto_expand
        == async_cisco_iosxe_conn._initialization_args.get("comms_auto_expand", False)
    )
    assert (
        async_cisco_iosxe_conn.channel.timeout_ops
        == async_cisco_iosxe_conn._initialization_args["timeout_ops"]
    )


@pytest.mark.asyncio
async def test_read_chunk(async_cisco_iosxe_conn):
    await async_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    async_cisco_iosxe_conn.transport.write(chunk)
    async_cisco_iosxe_conn.transport.write(async_cisco_iosxe_conn.channel.comms_return_char)
    # keep reading chunks until we have all the input data
    read_chunk = b""
    while chunk.encode() not in read_chunk:
        read_chunk += await async_cisco_iosxe_conn.channel._read_chunk()


@pytest.mark.asyncio
async def test_read_until_input(async_cisco_iosxe_conn):
    await async_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    async_cisco_iosxe_conn.transport.write(chunk)
    async_cisco_iosxe_conn.transport.write(async_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = await async_cisco_iosxe_conn.channel._read_until_input(
        channel_input=chunk.encode()
    )
    assert chunk.encode() in read_chunk


@pytest.mark.asyncio
async def test_read_until_prompt(async_cisco_iosxe_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_read_until_prompt"]["expected_prompt"]

    await async_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    async_cisco_iosxe_conn.transport.write(chunk)
    async_cisco_iosxe_conn.transport.write(async_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = await async_cisco_iosxe_conn.channel._read_until_input(
        channel_input=chunk.encode()
    )
    found_prompt = await async_cisco_iosxe_conn.get_prompt()
    assert chunk.encode() in read_chunk
    assert found_prompt == expected_prompt


@pytest.mark.asyncio
async def test_read_until_prompt_or_time(async_cisco_iosxe_conn):
    await async_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    async_cisco_iosxe_conn.transport.write(chunk)
    async_cisco_iosxe_conn.transport.write(async_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = await async_cisco_iosxe_conn.channel._read_until_prompt_or_time(
        channel_outputs=[b"chunk to read"]
    )
    assert chunk.encode() in read_chunk


@pytest.mark.asyncio
async def test_get_prompt(async_cisco_iosxe_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    await async_cisco_iosxe_conn.open()
    found_prompt = await async_cisco_iosxe_conn.get_prompt()
    assert found_prompt == expected_prompt


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
async def test_send_input(async_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    await async_cisco_iosxe_conn.open()
    raw_result, processed_result = await async_cisco_iosxe_conn.channel.send_input(
        channel_input="show version", strip_prompt=strip_prompt
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
async def test_send_input_and_read(async_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    await async_cisco_iosxe_conn.open()
    raw_result, processed_result = await async_cisco_iosxe_conn.channel.send_input_and_read(
        channel_input="show version", strip_prompt=strip_prompt
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()


@pytest.mark.asyncio
async def test_send_inputs_interact(async_cisco_iosxe_conn):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["raw_result"]
    expected_processed = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["processed_result"]
    interact_events = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["interact_events"]
    await async_cisco_iosxe_conn.open()
    raw_result, processed_result = await async_cisco_iosxe_conn.channel.send_inputs_interact(
        interact_events=interact_events
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()



"""scrapli.driver.base_generic_driver"""
from typing import List, Optional, Tuple, Union

from scrapli.helper import resolve_file
from scrapli.response import MultiResponse, Response


class GenericDriverBase:
    """
    GenericDriverBase Object

    A generic network driver that will *hopefully* work for a broad variety of devices with
    minimal to no modifications and provide a normal NetworkDriver type experience with
    `send_command(s)`, `get_prompt` and `send_interactive` methods instead of forcing users to
    call Channel methods directly.

    This driver doesn't know anything about privilege levels (or any type of "config modes",
    disabling paging, gracefully exiting, or anything like that, and as such should be treated
    similar to the base `Scrape` object from that perspective.

    """

    @staticmethod
    def _pre_send_command(
        host: str, command: str, failed_when_contains: Optional[Union[str, List[str]]] = None
    ) -> Response:
        """
        Handle pre "send_command" tasks for consistency between sync/async versions

        Args:
            host: string name of the host
            command: string to send to device in privilege exec mode
            failed_when_contains: string or list of strings indicating failure if found in response

        Returns:
            Response: Scrapli Response object

        Raises:
            TypeError: if command is anything but a string

        """
        if not isinstance(command, str):
            raise TypeError(
                f"`send_command` expects a single string, got {type(command)}, "
                "to send a list of commands use the `send_commands` method instead."
            )

        response = Response(
            host=host,
            channel_input=command,
            failed_when_contains=failed_when_contains,
        )

        return response

    @staticmethod
    def _post_send_command(
        raw_response: bytes, processed_response: bytes, response: Response
    ) -> Response:
        """
        Handle post "send_command" tasks for consistency between sync/async versions

        Args:
            raw_response: raw response returned from the channel
            processed_response: processed response returned from the channel
            response: response object to update with channel results

        Returns:
            Response: Scrapli Response object

        Raises:
            N/A

        """
        response._record_response(result=processed_response)  # pylint: disable=W0212
        response.raw_result = raw_response
        return response

    @staticmethod
    def _pre_send_commands(commands: List[str]) -> MultiResponse:
        """
        Handle pre "send_command" tasks for consistency between sync/async versions

        Args:
            commands: list of strings to send to device in privilege exec mode

        Returns:
            MultiResponse: Scrapli MultiResponse object

        Raises:
            TypeError: if command is anything but a string

        """
        if not isinstance(commands, list):
            raise TypeError(
                f"`send_commands` expects a list of strings, got {type(commands)}, "
                "to send a single command use the `send_command` method instead."
            )

        responses = MultiResponse()

        return responses

    @staticmethod
    def _pre_send_commands_from_file(file: str) -> List[str]:
        """
        Handle pre "send_commands_from_file" tasks for consistency between sync/async versions

        Args:
            file: string path to file

        Returns:
            commands: list of commands read from file

        Raises:
            TypeError: if anything but a string is provided for `file`

        """
        if not isinstance(file, str):
            raise TypeError(
                f"`send_commands_from_file` expects a string path to a file, got {type(file)}"
            )
        resolved_file = resolve_file(file)

        with open(resolved_file, "r") as f:
            commands = f.read().splitlines()

        return commands

    @classmethod
    def _pre_send_interactive(
        cls,
        host: str,
        interact_events: List[Tuple[str, str, Optional[bool]]],
        failed_when_contains: Optional[Union[str, List[str]]] = None,
    ) -> Response:
        """
        Handle pre "send_interactive" tasks for consistency between sync/async versions

        Args:
            host: string name of the host
            interact_events: list of tuples containing the "interactions" with the device
                each list element must have an input and an expected response, and may have an
                optional bool for the third and final element -- the optional bool specifies if the
                input that is sent to the device is "hidden" (ex: password), if the hidden param is
                not provided it is assumed the input is "normal" (not hidden)
            failed_when_contains: string or list of strings indicating failure if found in response

        Returns:
            Response: Scrapli Response object

        Raises:
            N/A

        """
        joined_input = ", ".join([event[0] for event in interact_events])
        return cls._pre_send_command(
            host=host, command=joined_input, failed_when_contains=failed_when_contains
        )



"""scrapli.driver.core"""
from scrapli.driver.core.arista_eos.async_driver import AsyncEOSDriver
from scrapli.driver.core.arista_eos.driver import EOSDriver
from scrapli.driver.core.cisco_iosxe.async_driver import AsyncIOSXEDriver
from scrapli.driver.core.cisco_iosxe.driver import IOSXEDriver
from scrapli.driver.core.cisco_iosxr.async_driver import AsyncIOSXRDriver
from scrapli.driver.core.cisco_iosxr.driver import IOSXRDriver
from scrapli.driver.core.cisco_nxos.async_driver import AsyncNXOSDriver
from scrapli.driver.core.cisco_nxos.driver import NXOSDriver
from scrapli.driver.core.juniper_junos.async_driver import AsyncJunosDriver
from scrapli.driver.core.juniper_junos.driver import JunosDriver

__all__ = (
    "AsyncEOSDriver",
    "EOSDriver",
    "AsyncIOSXEDriver",
    "IOSXEDriver",
    "AsyncIOSXRDriver",
    "IOSXRDriver",
    "AsyncNXOSDriver",
    "NXOSDriver",
    "AsyncJunosDriver",
    "JunosDriver",
)



def test_configuration_exclusive(iosxr_conn):
    iosxr_conn.acquire_priv("configuration_exclusive")
    result = iosxr_conn.send_configs(["do show configuration sessions"])
    iosxr_conn.close()
    # last character should be an asterisk indicating configuration is locked
    assert result[0].result[-1:] == "*"



import sys

import pytest

from scrapli.exceptions import ConnectionNotOpened, ScrapliAuthenticationFailed
from scrapli.transport import TelnetTransport
from scrapli.transport.telnet import ScrapliTelnet


def test_scrapli_telnet_type():
    # connection error will get raised because 23 shouldnt be open for most devices running the test
    with pytest.raises(ConnectionError):
        conn = ScrapliTelnet("localhost", 23, 100)
        assert conn.host == "localhost"
        assert conn.port == 23
        assert conn.timeout == 100


def test_creation():
    conn = TelnetTransport("localhost")
    assert conn.host == "localhost"
    assert conn.port == 23
    assert conn._isauthenticated is False


@pytest.mark.skipif(sys.platform.startswith("win"), reason="skipping on windows")
def test_open_failure():
    conn = TelnetTransport("localhost", port=23)
    with pytest.raises(ScrapliAuthenticationFailed) as exc:
        conn.open()
    assert str(exc.value) == "Failed to open telnet session to host localhost, connection refused"


def test_keepalive_standard():
    conn = TelnetTransport("localhost")
    with pytest.raises(NotImplementedError) as exc:
        conn._keepalive_standard()
    assert str(exc.value) == "No 'standard' keepalive mechanism for telnet."


@pytest.mark.parametrize(
    "method_name",
    ["read", "write", "set_timeout"],
    ids=["read", "write", "set_timeout"],
)
def test_requires_open(method_name):
    conn = TelnetTransport("localhost")
    method = getattr(conn, method_name)
    with pytest.raises(ConnectionNotOpened):
        if method_name == "write":
            method("blah")
        else:
            method()



"""scrapli.helper"""
import importlib
import os
import re
import warnings
from functools import lru_cache
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, TextIO, Tuple, Union

import pkg_resources  # pylint: disable=C0411

from scrapli.exceptions import TransportPluginError


def _find_transport_plugin(transport: str) -> Tuple[Any, Tuple[str, ...]]:
    """
    Find non-core transport plugins and required plugin arguments

    Args:
        transport: string name of the desired transport, i.e.: paramiko or ssh2

    Returns:
        transport_class: class representing the given transport
        required_transport_args: tuple of required arguments for given transport

    Raises:
        ModuleNotFoundError: if unable to  find scrapli transport module
        TransportPluginError: if unable to load `Transport` and `TRANSPORT_ARGS` from given
            transport module

    """
    try:
        transport_plugin_lib = importlib.import_module(f"scrapli_{transport}.transport")
    except ModuleNotFoundError as exc:
        err = f"Module '{exc.name}' not found!"
        msg = f"***** {err} {'*' * (80 - len(err))}"
        fix = (
            "To resolve this issue, ensure you are referencing a valid transport plugin. Transport"
            " plugins should be named similar to `scrapli_paramiko` or `scrapli_ssh2`, and can be "
            "selected by passing simply `paramiko` or `ssh2` into the scrapli driver. You can "
            "install most plugins with pip: `pip install scrapli-ssh2` for example."
        )
        warning = "\n" + msg + "\n" + fix + "\n" + msg
        raise ModuleNotFoundError(warning) from exc
    transport_class = getattr(transport_plugin_lib, "Transport", None)
    required_transport_args = getattr(transport_plugin_lib, "TRANSPORT_ARGS", None)
    if not all([transport_class, required_transport_args]):
        msg = f"Failed to load transport plugin `{transport}` transport class or required arguments"
        raise TransportPluginError(msg)
    return transport_class, required_transport_args


@lru_cache()
def get_prompt_pattern(prompt: str, class_prompt: str) -> Pattern[bytes]:
    """
    Return compiled prompt pattern

    Given a potential prompt and the Channel class' prompt, return compiled prompt pattern

    Args:
        prompt: bytes string to process
        class_prompt: Channel class prompt pattern; never re.escape class prompt pattern

    Returns:
        output: bytes string each line right stripped

    Raises:
        N/A

    """
    check_prompt = prompt or class_prompt
    if isinstance(check_prompt, str):
        bytes_check_prompt = check_prompt.encode()
    else:
        bytes_check_prompt = check_prompt

    if bytes_check_prompt.startswith(b"^") and bytes_check_prompt.endswith(b"$"):
        return re.compile(bytes_check_prompt, flags=re.M | re.I)
    if check_prompt == class_prompt:
        return re.compile(bytes_check_prompt, flags=re.M | re.I)
    return re.compile(re.escape(bytes_check_prompt))


def normalize_lines(output: bytes) -> bytes:
    r"""
    Normalize lines

    Split output lines to remove \r\n, rstrip each line and rejoin

    Args:
        output: bytes string to process

    Returns:
        bytes: bytes string each line right stripped

    Raises:
        N/A

    """
    return b"\n".join([line.rstrip() for line in output.splitlines()])


def strip_ansi(output: bytes) -> bytes:
    """
    Strip ansi characters from output

    Args:
        output: bytes from previous reads if needed

    Returns:
        bytes: bytes output read from channel with ansi characters removed

    Raises:
        N/A

    """
    ansi_escape_pattern = re.compile(rb"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))")
    output = re.sub(pattern=ansi_escape_pattern, repl=b"", string=output)
    return output


def _textfsm_get_template(platform: str, command: str) -> Optional[TextIO]:
    """
    Find correct TextFSM template based on platform and command executed

    Args:
        platform: ntc-templates device type; i.e. cisco_ios, arista_eos, etc.
        command: string of command that was executed (to find appropriate template)

    Returns:
        None or TextIO of opened template

    Raises:
        N/A

    """
    try:
        from ntc_templates import templates  # pylint: disable=C0415,W0611
        from textfsm.clitable import CliTable  # pylint: disable=C0415
    except ModuleNotFoundError as exc:
        err = f"Module '{exc.name}' not installed!"
        msg = f"***** {err} {'*' * (80 - len(err))}"
        fix = (
            f"To resolve this issue, install '{exc.name}'. You can do this in one of the following"
            " ways:\n"
            "1: 'pip install -r requirements-textfsm.txt'\n"
            "2: 'pip install scrapli[textfsm]'"
        )
        warning = "\n" + msg + "\n" + fix + "\n" + msg
        warnings.warn(warning)
        return None
    template_dir = pkg_resources.resource_filename("ntc_templates", "templates")
    cli_table = CliTable("index", template_dir)
    template_index = cli_table.index.GetRowMatch({"Platform": platform, "Command": command})
    if not template_index:
        return None
    template_name = cli_table.index.index[template_index]["Template"]
    template = open(f"{template_dir}/{template_name}")
    return template


def _textfsm_to_dict(
    structured_output: Union[List[Any], Dict[str, Any]], header: List[str]
) -> Union[List[Any], Dict[str, Any]]:
    """
    Create list of dicts from textfsm output and header

    Args:
        structured_output: parsed textfsm output
        header: list of strings representing column headers for textfsm output

    Returns:
        output: structured data

    Raises:
        N/A

    """
    header_lower = [h.lower() for h in header]
    structured_output = [dict(zip(header_lower, row)) for row in structured_output]
    return structured_output


def textfsm_parse(
    template: Union[str, TextIOWrapper], output: str, to_dict: bool = True
) -> Union[List[Any], Dict[str, Any]]:
    """
    Parse output with TextFSM and ntc-templates, try to return structured output

    Args:
        template: TextIOWrapper or string path to template to use to parse data
        output: unstructured output from device to parse
        to_dict: convert textfsm output from list of lists to list of dicts -- basically create dict
            from header and row data so it is easier to read/parse the output

    Returns:
        output: structured data

    Raises:
        N/A

    """
    import textfsm  # pylint: disable=C0415

    if not isinstance(template, TextIOWrapper):
        template_file = open(template)
    else:
        template_file = template
    re_table = textfsm.TextFSM(template_file)
    try:
        structured_output: Union[List[Any], Dict[str, Any]] = re_table.ParseText(output)
        if to_dict:
            structured_output = _textfsm_to_dict(
                structured_output=structured_output, header=re_table.header
            )
        return structured_output
    except textfsm.parser.TextFSMError:
        pass
    return []


def genie_parse(platform: str, command: str, output: str) -> Union[List[Any], Dict[str, Any]]:
    """
    Parse output with Cisco genie parsers, try to return structured output

    Args:
        platform: genie device type; i.e. iosxe, iosxr, etc.
        command: string of command that was executed (to find appropriate parser)
        output: unstructured output from device to parse

    Returns:
        output: structured data

    Raises:
        N/A

    """
    try:
        from genie.conf.base import Device  # pylint: disable=C0415
        from genie.libs.parser.utils import get_parser  # pylint: disable=C0415
    except ModuleNotFoundError as exc:
        err = f"Module '{exc.name}' not installed!"
        msg = f"***** {err} {'*' * (80 - len(err))}"
        fix = (
            f"To resolve this issue, install '{exc.name}'. You can do this in one of the following"
            " ways:\n"
            "1: 'pip install -r requirements-genie.txt'\n"
            "2: 'pip install scrapli[genie]'"
        )
        warning = "\n" + msg + "\n" + fix + "\n" + msg
        warnings.warn(warning)
        return []

    genie_device = Device("scrapli_device", custom={"abstraction": {"order": ["os"]}}, os=platform)

    try:
        get_parser(command, genie_device)
        genie_parsed_result = genie_device.parse(command, output=output)
        if not genie_parsed_result:
            return []
        if isinstance(genie_parsed_result, (list, dict)):
            return genie_parsed_result
    # have to catch base exception because that is all genie raises for some reason :(
    except Exception:  # pylint: disable=E0012,W0703
        pass
    return []


def resolve_ssh_config(ssh_config_file: str) -> str:
    """
    Resolve ssh configuration file from provided string

    If provided string is empty (`""`) try to resolve system ssh config files located at
    `~/.ssh/config` or `/etc/ssh/ssh_config`.

    Args:
        ssh_config_file: string representation of ssh config file to try to use

    Returns:
        str: string path to ssh config file or an empty string

    Raises:
        N/A

    """
    if Path(ssh_config_file).is_file():
        return str(Path(ssh_config_file))
    if Path(os.path.expanduser("~/.ssh/config")).is_file():
        return str(Path(os.path.expanduser("~/.ssh/config")))
    if Path("/etc/ssh/ssh_config").is_file():
        return str(Path("/etc/ssh/ssh_config"))
    return ""


def resolve_ssh_known_hosts(ssh_known_hosts: str) -> str:
    """
    Resolve ssh known hosts file from provided string

    If provided string is empty (`""`) try to resolve system known hosts files located at
    `~/.ssh/known_hosts` or `/etc/ssh/ssh_known_hosts`.

    Args:
        ssh_known_hosts: string representation of ssh config file to try to use

    Returns:
        str: string path to ssh known hosts file or an empty string

    Raises:
        N/A

    """
    if Path(ssh_known_hosts).is_file():
        return str(Path(ssh_known_hosts))
    if Path(os.path.expanduser("~/.ssh/known_hosts")).is_file():
        return str(Path(os.path.expanduser("~/.ssh/known_hosts")))
    if Path("/etc/ssh/ssh_known_hosts").is_file():
        return str(Path("/etc/ssh/ssh_known_hosts"))
    return ""


def resolve_file(file: str) -> str:
    """
    Resolve file from provided string

    Args:
        file: string path to file

    Returns:
        str: string path to file

    Raises:
        ValueError: if file cannot be resolved

    """
    if Path(file).is_file():
        return str(Path(file))
    if Path(os.path.expanduser(file)).is_file():
        return str(Path(os.path.expanduser(file)))
    raise ValueError(f"File path `{file}` could not be resolved")



from pathlib import Path

import pytest

import scrapli
from scrapli.response import MultiResponse, Response

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data/"


def test_pre_send_command_exception(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_command(
            host=sync_cisco_iosxe_conn.transport.host, command=[]
        )
    assert (
        str(exc.value)
        == "`send_command` expects a single string, got <class 'list'>, to send a list of commands use the `send_commands` method instead."
    )


def test_pre_send_command(sync_cisco_iosxe_conn):
    response = sync_cisco_iosxe_conn._pre_send_command(
        host=sync_cisco_iosxe_conn.transport.host,
        command="show version",
        failed_when_contains=["something"],
    )
    assert isinstance(response, Response)
    assert response.host == sync_cisco_iosxe_conn.transport.host
    assert response.failed_when_contains == ["something"]
    assert response.channel_input == "show version"


def test_post_send_command(sync_cisco_iosxe_conn):
    response = sync_cisco_iosxe_conn._pre_send_command(
        host=sync_cisco_iosxe_conn.transport.host,
        command="show version",
        failed_when_contains=["something"],
    )
    final_response = sync_cisco_iosxe_conn._post_send_command(
        raw_response=b"blah", processed_response=b"blahx2", response=response
    )
    # generic driver doesnt know/care about textfsm_platform
    assert final_response.textfsm_platform == ""
    assert final_response.start_time == response.start_time
    assert final_response.elapsed_time > 0


def test_pre_send_commands_exception(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_commands(commands="boo")
    assert (
        str(exc.value)
        == "`send_commands` expects a list of strings, got <class 'str'>, to send a single command use the `send_command` method instead."
    )


def test_pre_send_commands(sync_cisco_iosxe_conn):
    responses = sync_cisco_iosxe_conn._pre_send_commands(commands=["show version"])
    assert isinstance(responses, MultiResponse)
    assert responses.failed is False


def test_pre_send_commands_from_file_exception(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn._pre_send_commands_from_file(file=[])
    assert (
        str(exc.value)
        == "`send_commands_from_file` expects a string path to a file, got <class 'list'>"
    )


def test_pre_send_commands_from_file(sync_cisco_iosxe_conn):
    commands = sync_cisco_iosxe_conn._pre_send_commands_from_file(
        file=f"{TEST_DATA_DIR}/files/vrnetlab_key"
    )
    assert commands[0] == "-----BEGIN OPENSSH PRIVATE KEY-----"
    assert commands[-1] == "-----END OPENSSH PRIVATE KEY-----"


def test_pre_send_commands_interactive(sync_cisco_iosxe_conn):
    response = sync_cisco_iosxe_conn._pre_send_interactive(
        host=sync_cisco_iosxe_conn.transport.host,
        interact_events=[("input1", "expected1"), ("input2", "expected2")],
    )
    assert isinstance(response, Response)
    assert response.host == sync_cisco_iosxe_conn.transport.host
    assert response.failed_when_contains is None
    assert response.channel_input == "input1, input2"



"""scrapli.driver.core.cisco_iosxr"""
from scrapli.driver.core.cisco_iosxr.async_driver import AsyncIOSXRDriver
from scrapli.driver.core.cisco_iosxr.driver import IOSXRDriver

__all__ = (
    "AsyncIOSXRDriver",
    "IOSXRDriver",
)



from scrapli.driver.core import NXOSDriver


def test_nxos_driver_init_telnet():
    conn = NXOSDriver(host="myhost", transport="telnet")
    assert conn.transport.username_prompt == "login:"


def test_nxos_open_close(sync_cisco_nxos_conn):
    sync_cisco_nxos_conn.open()
    assert sync_cisco_nxos_conn.get_prompt() == "switch#"
    # "more" will show up if we haven't sent terminal length 0 to the mock nxos server
    assert "more" not in sync_cisco_nxos_conn.send_command(command="show version").result.lower()
    sync_cisco_nxos_conn.close()



"""scrapli.driver.driver"""
from types import TracebackType
from typing import Any, Optional, Type

from scrapli.channel import Channel
from scrapli.driver.base_driver import ASYNCIO_TRANSPORTS, ScrapeBase
from scrapli.exceptions import TransportPluginError


class Scrape(ScrapeBase):
    def __init__(self, **kwargs: Any) -> None:
        """
        Scrape Object

        Scrape is the base class for NetworkDriver, and subsequent platform specific drivers (i.e.
        IOSXEDriver). Scrape can be used on its own and offers a semi-pexpect like experience in
        that it doesn't know or care about privilege levels, platform types, and things like that.

        *Note* most arguments passed to Scrape do not actually get assigned to the scrape object
        itself, but instead are used to construct the Transport and Channel classes that Scrape
        relies on, see Transport and Channel docs for details.

        Args:
            kwargs: Keyword arguments to pass to `ScrapeBase` -- see `ScrapeBase` for available args

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TransportPluginError: if attempting to use an asyncio transport plugin

        """
        super().__init__(**kwargs)

        if self._transport in ASYNCIO_TRANSPORTS:
            raise TransportPluginError(
                f"Attempting to use transport type {self._transport} with a sync driver, "
                "must use a non-asyncio transport"
            )

        self.channel = Channel(transport=self.transport, **self.channel_args)

    def __enter__(self) -> "Scrape":
        """
        Enter method for context manager

        Args:
            N/A

        Returns:
            self: instance of self

        Raises:
            N/A

        """
        self.open()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Exit method to cleanup for context manager

        Args:
            exception_type: exception type being raised
            exception_value: message from exception being raised
            traceback: traceback from exception being raised

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.close()

    def open(self) -> None:
        """
        Open Transport (socket/session) and establish channel

        If on_open callable provided, execute that callable after opening connection

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger.info(f"Opening connection to {self._initialization_args['host']}")
        self.transport.open()
        if self.on_open:
            self.on_open(self)
        self.logger.info(f"Connection to {self._initialization_args['host']} opened successfully")

    def close(self) -> None:
        """
        Close Transport (socket/session)

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.logger.info(f"Closing connection to {self._initialization_args['host']}")
        if self.on_close:
            self.on_close(self)
        self.transport.close()
        self.logger.info(f"Connection to {self._initialization_args['host']} closed successfully")

    @property
    def timeout_ops(self) -> float:
        """
        Property for timeout_ops attribute

        Args:
            N/A

        Returns:
            float: value of timeout_ops

        Raises:
            N/A

        """
        return self._timeout_ops

    @timeout_ops.setter
    def timeout_ops(self, timeout_value: float) -> None:
        """
        Setter for timeout_ops attribute

        Sets timeout value for the channel class as well as the internal `_timeout_ops` value of the
        base driver class

        Args:
            timeout_value: float value to set as timeout_ops

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.channel.timeout_ops = timeout_value
        self._timeout_ops = timeout_value



import os
import re
import sys
from io import TextIOWrapper
from pathlib import Path

import pkg_resources  # pylint: disable=C041
import pytest

import scrapli
from scrapli.helper import (
    _textfsm_get_template,
    genie_parse,
    get_prompt_pattern,
    resolve_file,
    strip_ansi,
    textfsm_parse,
)

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"

IOS_ARP = """Protocol  Address          Age (min)  Hardware Addr   Type   Interface
Internet  172.31.254.1            -   0000.0c07.acfe  ARPA   Vlan254
Internet  172.31.254.2            -   c800.84b2.e9c2  ARPA   Vlan254
"""


def test_get_prompt_pattern_class_pattern():
    class_pattern = "^averygoodpattern$"
    result = get_prompt_pattern("", class_pattern)
    assert result == re.compile(b"^averygoodpattern$", re.IGNORECASE | re.MULTILINE)


def test_get_prompt_pattern_class_pattern_no_line_start_end_markers():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern(class_pattern, "")
    assert result == re.compile(b"averygoodpattern")


def test_get_prompt_pattern_arg_pattern():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern("^awesomepattern$", class_pattern)
    assert result == re.compile(b"^awesomepattern$", re.IGNORECASE | re.MULTILINE)


def test_get_prompt_pattern_arg_string():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern("awesomepattern", class_pattern)
    assert result == re.compile(b"awesomepattern")


def test__strip_ansi():
    output = b"[admin@CoolDevice.Sea1: \x1b[1m/\x1b[0;0m]$"
    output = strip_ansi(output)
    assert output == b"[admin@CoolDevice.Sea1: /]$"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test__textfsm_get_template_valid_template():
    template = _textfsm_get_template("cisco_nxos", "show ip arp")
    template_dir = pkg_resources.resource_filename("ntc_templates", "templates")
    assert isinstance(template, TextIOWrapper)
    assert template.name == f"{template_dir}/cisco_nxos_show_ip_arp.textfsm"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test__textfsm_get_template_invalid_template():
    template = _textfsm_get_template("cisco_nxos", "show racecar")
    assert not template


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
@pytest.mark.parametrize(
    "parse_type",
    [
        (
            False,
            ["Internet", "172.31.254.1", "-", "0000.0c07.acfe", "ARPA", "Vlan254"],
        ),
        (
            True,
            {
                "protocol": "Internet",
                "address": "172.31.254.1",
                "age": "-",
                "mac": "0000.0c07.acfe",
                "type": "ARPA",
                "interface": "Vlan254",
            },
        ),
    ],
    ids=["to_dict_false", "to_dict_true"],
)
def test_textfsm_parse_success(parse_type):
    to_dict = parse_type[0]
    expected_result = parse_type[1]
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template, IOS_ARP, to_dict=to_dict)
    assert isinstance(result, list)
    assert result[0] == expected_result


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
@pytest.mark.parametrize(
    "parse_type",
    [
        (
            False,
            ["Internet", "172.31.254.1", "-", "0000.0c07.acfe", "ARPA", "Vlan254"],
        ),
        (
            True,
            {
                "protocol": "Internet",
                "address": "172.31.254.1",
                "age": "-",
                "mac": "0000.0c07.acfe",
                "type": "ARPA",
                "interface": "Vlan254",
            },
        ),
    ],
    ids=["to_dict_false", "to_dict_true"],
)
def test_textfsm_parse_success_string_path(parse_type):
    to_dict = parse_type[0]
    expected_result = parse_type[1]
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template.name, IOS_ARP, to_dict=to_dict)
    assert isinstance(result, list)
    assert result[0] == expected_result


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test_textfsm_parse_failure():
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template, "not really arp data")
    assert result == []


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting genie on windows")
def test_genie_parse_success():
    result = genie_parse("iosxe", "show ip arp", IOS_ARP)
    assert isinstance(result, dict)
    assert (
        result["interfaces"]["Vlan254"]["ipv4"]["neighbors"]["172.31.254.1"]["ip"] == "172.31.254.1"
    )


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting genie on windows")
def test_genie_parse_failure():
    result = genie_parse("iosxe", "show ip arp", "not really arp data")
    assert result == []
    # genie loads about nine million modules... for whatever reason these two upset pyfakefs
    del sys.modules["ats.configuration"]
    del sys.modules["pyats.configuration"]


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not dealing with windows path things in tests"
)
def test_resolve_file():
    resolved_file = resolve_file(file=f"{TEST_DATA_DIR}/files/_ssh_config")
    assert resolved_file == f"{TEST_DATA_DIR}/files/_ssh_config"


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not dealing with windows path things in tests"
)
def test_resolve_file_expanduser(fs):
    fs.add_real_file(
        source_path=f"{TEST_DATA_DIR}/files/_ssh_config",
        target_path=f"{os.path.expanduser('~')}/myneatfile",
    )
    resolved_file = resolve_file(file=f"~/myneatfile")
    assert resolved_file == f"{os.path.expanduser('~')}/myneatfile"


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not dealing with windows path things in tests"
)
def test_resolve_file_failure():
    with pytest.raises(ValueError) as exc:
        resolve_file(file=f"~/myneatfile")
    assert str(exc.value) == "File path `~/myneatfile` could not be resolved"



import pytest

from ...test_data.unit_test_cases import TEST_CASES


def test__str(sync_cisco_iosxe_conn):
    assert str(sync_cisco_iosxe_conn.channel) == "scrapli Channel Object"


def test__repr(sync_cisco_iosxe_conn):
    assert (
        repr(sync_cisco_iosxe_conn.channel)
        == "scrapli Channel {'logger': 'scrapli.channel-localhost', 'comms_prompt_pattern': '(^[a-z0-9.\\\\-_@()/:]{1,"
        "63}>$)|(^[a-z0-9.\\\\-_@/:]{1,63}#$)|(^[a-z0-9.\\\\-_@/:]{1,63}\\\\(conf[a-z0-9.\\\\-@/:\\\\+]{0,32}\\\\)#$)', "
        "'comms_return_char': '\\n', 'comms_ansi': True, 'comms_auto_expand': False, 'timeout_ops': 30.0}"
    )


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({"strip_prompt": True}, b"hostname 3560CX"),
        ({"strip_prompt": False}, b"hostname 3560CX\n3560CX#"),
    ],
    ids=[
        "strip_prompt_true",
        "strip_prompt_false",
    ],
)
def test__restructure_output(sync_cisco_iosxe_conn, attr_setup):
    args = attr_setup[0]
    expected = attr_setup[1]
    channel_output = b"hostname 3560CX\r\n3560CX#"
    output = sync_cisco_iosxe_conn.channel._restructure_output(channel_output, **args)
    assert output == expected


def test_pre_send_input(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn.channel._pre_send_input(channel_input=[])
    assert str(exc.value) == "`send_input` expects a single string, got <class 'list'>."


def test_pre_send_inputs_interact(sync_cisco_iosxe_conn):
    with pytest.raises(TypeError) as exc:
        sync_cisco_iosxe_conn.channel._pre_send_inputs_interact(interact_events="blah")
    assert str(exc.value) == "`interact_events` expects a List, got <class 'str'>"


def test_post_send_inputs_interact(sync_cisco_iosxe_conn):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"][
        "no_strip"
    ]
    raw_result, processed_result = sync_cisco_iosxe_conn.channel._post_send_inputs_interact(
        output=expected_raw.encode()
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()


@pytest.mark.parametrize(
    "attr_setup",
    [(True, "show version", "sho ver"), (False, "show vkasjflkdsjafl", "sho ver")],
    ids=["not expanded", "expanded"],
)
def test_process_auto_expand(sync_cisco_iosxe_conn, attr_setup):
    expanded = attr_setup[0]
    output = attr_setup[1]
    channel_input = attr_setup[2]
    result = sync_cisco_iosxe_conn.channel._process_auto_expand(
        output=output, channel_input=channel_input
    )
    assert expanded == result



"""mock_devices.juniper_junos"""
from typing import Any

from .base import BaseServer, BaseSSHServerSession

# yes this is "show version detail" but w/e it requires paging :)
SHOW_COMMAND = """Model: vsrx
Junos: 17.3R2.10
JUNOS Software Release [17.3R2.10]
KERNEL 17.3R2.10 #0 built by builder on 2018-02-08 02:19:07 UTC
MGD release 17.3R2.10 built by builder on 2018-02-08 04:40:18 UTC
CLI release 17.3R2.10 built by builder on 2018-02-08 03:45:31 UTC
IKED release 17.3R2.10 built by builder on 2018-02-08 04:03:22 UTC
RPD release 17.3R2.10 built by builder on 2018-02-08 05:08:08 UTC
CHASSISD release 17.3R2.10 built by builder on 2018-02-08 04:41:34 UTC
GKSD release 17.3R2.10 built by builder on 2018-02-08 04:05:26 UTC
GKMD release 17.3R2.10 built by builder on 2018-02-08 04:04:37 UTC
PKID release 17.3R2.10 built by builder on 2018-02-08 03:56:43 UTC
SENDD release 17.3R2.10 built by builder on 2018-02-08 03:44:42 UTC
DFWD release 17.3R2.10 built by builder on 2018-02-08 03:58:56 UTC
SNMPD release 17.3R2.10 built by builder on 2018-02-08 04:37:26 UTC
MIB2D release 17.3R2.10 built by builder on 2018-02-08 04:40:40 UTC
VRRPD release 17.3R2.10 built by builder on 2018-02-08 04:40:45 UTC
ALARMD release 17.3R2.10 built by builder on 2018-02-08 03:58:13 UTC
PFED release 17.3R2.10 built by builder on 2018-02-08 04:40:40 UTC
CRAFTD release 17.3R2.10 built by builder on 2018-02-08 03:58:51 UTC
SAMPLED release 17.3R2.10 built by builder on 2018-02-08 03:59:57 UTC
SRRD release 17.3R2.10 built by builder on 2018-02-08 03:49:41 UTC
RMOPD release 17.3R2.10 built by builder on 2018-02-08 03:58:02 UTC
COSD release 17.3R2.10 built by builder on 2018-02-08 04:02:26 UTC
IRSD release 17.3R2.10 built by builder on 2018-02-08 03:49:21 UTC
FUD release 17.3R2.10 built by builder on 2018-02-08 03:46:48 UTC
KSYNCD release 17.3R2.10 built by builder on 2018-02-08 03:47:19 UTC
HTTPD-GK release 17.3R2.10 built by builder on 2018-02-08 03:52:07 UTC
DHCPD release 17.3R2.10 built by builder on 2018-02-08 03:50:41 UTC
PPPOED release 17.3R2.10 built by builder on 2018-02-08 04:41:39 UTC
PPPD release 17.3R2.10 built by builder on 2018-02-08 04:41:16 UTC
DFCD release 17.3R2.10 built by builder on 2018-02-08 03:58:55 UTC
LACPD release 17.3R2.10 built by builder on 2018-02-08 04:40:39 UTC
USBD release 17.3R2.10 built by builder on 2018-02-08 03:21:06 UTC
LFMD release 17.3R2.10 built by builder on 2018-02-08 03:59:45 UTC
OAMD release 17.3R2.10 built by builder on 2018-02-08 04:01:15 UTC
TNETD release 17.3R2.10 built by builder on 2018-02-08 03:21:12 UTC
CFMD release 17.3R2.10 built by builder on 2018-02-08 03:58:25 UTC
JDHCPD release 17.3R2.10 built by builder on 2018-02-08 04:40:34 UTC
PSSD release 17.3R2.10 built by builder on 2018-02-08 03:49:29 UTC
AUTHD release 17.3R2.10 built by builder on 2018-02-08 04:40:24 UTC
BDBREPD release 17.3R2.10 built by builder on 2018-02-08 03:46:18 UTC
RES-CLEANUPD release 17.3R2.10 built by builder on 2018-02-08 03:47:48 UTC
APPIDD release 17.3R2.10 built by builder on 2018-02-08 03:52:03 UTC
IDPD release 17.3R2.10 built by builder on 2018-02-08 04:03:21 UTC
SHM-RTSDBD release 17.3R2.10 built by builder on 2018-02-08 03:47:50 UTC
SMID release 17.3R2.10 built by builder on 2018-02-08 04:40:42 UTC
SMIHELPERD release 17.3R2.10 built by builder on 2018-02-08 04:00:00 UTC
R2CPD release 17.3R2.10 built by builder on 2018-02-08 03:49:35 UTC
GSTATD release 17.3R2.10 built by builder on 2018-02-08 03:44:42 UTC
DOT1XD release 17.3R2.10 built by builder on 2018-02-08 04:29:39 UTC
UACD release 17.3R2.10 built by builder on 2018-02-08 03:50:38 UTC
ESSMD release 17.3R2.10 built by builder on 2018-02-08 03:49:02 UTC
TAD release 17.3R2.10 built by builder on 2018-02-08 04:01:41 UTC
DOOD release 17.3R2.10 built by builder on 2018-02-08 03:51:17 UTC
AUTOD release 17.3R2.10 built by builder on 2018-02-08 03:48:15 UTC
NSD release 17.3R2.10 built by builder on 2018-02-08 04:04:20 UTC
IPFD release 17.3R2.10 built by builder on 2018-02-08 04:03:23 UTC
AAMWD release 17.3R2.10 built by builder on 2018-02-08 04:03:17 UTC
NSTRACED release 17.3R2.10 built by builder on 2018-02-08 03:51:00 UTC
FWAUTHD release 17.3R2.10 built by builder on 2018-02-08 04:04:06 UTC
GPRSD release 17.3R2.10 built by builder on 2018-02-08 03:50:56 UTC
JSRPD release 17.3R2.10 built by builder on 2018-02-08 04:04:13 UTC
PROFILERD release 17.3R2.10 built by builder on 2018-02-08 03:51:04 UTC
OAMD release 17.3R2.10 built by builder on 2018-02-08 04:01:15 UTC
RTLOGD release 17.3R2.10 built by builder on 2018-02-08 04:04:20 UTC
UTMD release 17.3R2.10 built by builder on 2018-02-08 04:04:34 UTC
SYSHMD release 17.3R2.10 built by builder on 2018-02-08 03:51:22 UTC
SMTPD release 17.3R2.10 built by builder on 2018-02-08 03:51:07 UTC
USERIDD release 17.3R2.10 built by builder on 2018-02-08 04:03:33 UTC
SDK-VMMD release 17.3R2.10 built by builder on 2018-02-08 04:00:53 UTC
PPMD release 17.3R2.10 built by builder on 2018-02-08 03:57:48 UTC
STATICD release 17.3R2.10 built by builder on 2018-02-08 04:39:50 UTC
PGMD release 17.3R2.10 built by builder on 2018-02-08 03:45:05 UTC
BFDD release 17.3R2.10 built by builder on 2018-02-08 03:57:07 UTC
AUDITD release 17.3R2.10 built by builder on 2018-02-08 03:44:47 UTC
L2ALD release 17.3R2.10 built by builder on 2018-02-08 04:38:48 UTC
EVENTD release 17.3R2.10 built by builder on 2018-02-08 03:57:09 UTC
L2CPD release 17.3R2.10 built by builder on 2018-02-08 04:38:54 UTC
MCSNOOPD release 17.3R2.10 built by builder on 2018-02-08 04:39:04 UTC
MPLSOAMD release 17.3R2.10 built by builder on 2018-02-08 03:44:57 UTC
COMMITD release 17.3R2.10 built by builder on 2018-02-08 04:40:13 UTC
JNUD release 17.3R2.10 built by builder on 2018-02-08 04:38:47 UTC
WEB-API release 17.3R2.10 built by builder on 2018-02-08 03:20:56 UTC
JSD release 17.3R2.10 built by builder on 2018-02-08 04:40:45 UTC
UI-PUBD release 17.3R2.10 built by builder on 2018-02-08 03:48:53 UTC
MGD-API release 17.3R2.10 built by builder on 2018-02-08 04:40:18 UTC
MGD-API release 17.3R2.10 built by builder on 2018-02-08 04:40:18 UTC
NTAD release 17.3R2.10 built by builder on 2018-02-08 03:45:01 UTC
SDPD release 17.3R2.10 built by builder on 2018-02-08 03:45:22 UTC
SPMD release 17.3R2.10 built by builder on 2018-02-08 03:57:52 UTC
SCPD release 17.3R2.10 built by builder on 2018-02-08 03:45:20 UTC
SDXD release 17.3R2.10 built by builder on 2018-02-08 03:44:42 UTC
base-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:19 UTC
jroute-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:47 UTC
jkernel-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:39 UTC
appsecure-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:34 UTC
aprobe-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:34 UTC
authd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:34 UTC
autod-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:06 UTC
cfm-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:35 UTC
chassis_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:35 UTC
collector-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
cos_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:36 UTC
cpcdd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:36 UTC
dcd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:37 UTC
dfcd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:37 UTC
dot1xd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:08 UTC
elmi-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:37 UTC
essmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
forwarding_options_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:37 UTC
gres-test-point-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:47 UTC
httpd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:38 UTC
iccp_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:39 UTC
jappid-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:39 UTC
jcrypto-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:33 UTC
jcrypto_usp-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:12 UTC
jdocs-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:32 UTC
jidpd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:39 UTC
jkernel_jseries-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:14 UTC
jkernel_usp-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:12 UTC
jpppd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
jroute_junos-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:47 UTC
l2ald-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:48 UTC
lldp-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:48 UTC
lrf-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
mclag_cfgchk_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:08 UTC
mcsnoop-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:49 UTC
mo-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:41 UTC
pccd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:49 UTC
phcd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:14 UTC
pppd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
pppoed-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
r2cpd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
rdd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:42 UTC
repd_cmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:58:42 UTC
scpd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:50 UTC
sdpd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:50 UTC
services-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:09 UTC
spmd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:56:53 UTC
srd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:10 UTC
stpng-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:07 UTC
syshmd_health_mon-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:14 UTC
syshmd_trackip-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:14 UTC
traffic-dird-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:10 UTC
transportd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:10 UTC
url-filterd-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:10 UTC
vlans-ng-actions-dd release 17.3R2.10 built by builder on 2018-02-08 02:59:07 UTC

"""


class JunoSSHServerSession(BaseSSHServerSession):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Juniper Junos mock ssh server

        Args:
            args: args to pass to base class
            kwargs: keyword args to pass to base class

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(*args, **kwargs)

        self.terminal_length = 32

        self.priv_levels = {
            "exec": "vrnetlab>",
            "config": "vrnetlab#",
        }
        self.priv_level = "exec"
        self.invalid_input = "% Invalid command at '^' marker.\n"
        self.paging_prompt = "\n--(more)--"

        self.command_mapper = {
            self.handle_open: [
                "set cli complete-on-space off",
                "set cli screen-length 0",
                "set cli screen-width 511",
            ],
            self.handle_close: ["exit"],
            self.handle_priv: ["configure", "exit"],
            self.handle_show: ["show version"],
            self.handle_interactive: [],
        }

        self.show_command_output = SHOW_COMMAND

    def handle_open(self, channel_input: str) -> None:
        """
        Handle "open" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "config":
            self.handle_invalid(channel_input=channel_input)
        elif channel_input == "set cli screen-length 0":
            # in the future can actually set the value if we care?
            self.terminal_length = 0
            self._chan.write("")
            self.repaint_prompt()
        elif channel_input == "set cli screen-width 511":
            self.terminal_width = 511
            self._chan.write("")
            self.repaint_prompt()
        elif channel_input == "set cli complete-on-space off":
            # we dont care about this really, but scrapli will send it during open so have to just
            # make sure we accept it and move on
            self._chan.write("")
            self.repaint_prompt()
        else:
            self.handle_invalid(channel_input=channel_input)

    def handle_close(self, channel_input: str) -> None:
        """
        Handle "close" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = channel_input
        if self.priv_level == "config":
            # exit is a weird situation since it can be used to exit config mode or exit the device
            # so this lives here instead of handle_priv
            self._chan.write("")
            self.priv_level = "exec"
            self.repaint_prompt()
        else:
            self.eof_received()

    def _handle_priv_configure(self, channel_input: str) -> None:
        """
        Handle the "configure terminal" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level != "exec":
            self.handle_invalid(channel_input=channel_input)
        else:
            self._chan.write("")
            self.priv_level = "config"
            self.repaint_prompt()

    def handle_priv(self, channel_input: str) -> None:
        """
        Handle "privilege" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        # junos has just "exit", no "end" and no "disable/enable" so just configure to deal w/ here
        if channel_input == "configure":
            self._handle_priv_configure(channel_input=channel_input)

    def handle_interactive(self, channel_input: str) -> None:
        """
        Handle the "interactive" (clear logging) command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        # TODO
        pass

    def data_received(self, data: str, datatype: None) -> None:
        """
        Handle data received on ssh channel

        Args:
            data: string of data sent to channel
            datatype: dunno! in base class though :)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        # in the future we can cutoff the inputs if it is over X width if disable width has not yet
        # been ran -- not needed now but could be cool; if we just send a return, we should NOT
        # strip that!
        channel_input = data if data == "\n" else data.rstrip()

        if data == "\n":
            self.repaint_prompt()
            return

        for handler, channel_inputs in self.command_mapper.items():
            if channel_input in channel_inputs:
                # as soon as we find a match for the channel input ship it to the handler; handler
                # can figure out if command is valid for the given priv and such
                handler(channel_input=channel_input)
                return

        self.handle_invalid(channel_input=channel_input)


class JunosServer(BaseServer):
    pass



import re


def cisco_iosxe_clean_response(response):
    def _replace_config_bytes(response):
        config_bytes_pattern = re.compile(r"^Current configuration : \d+ bytes$", flags=re.M | re.I)
        response = re.sub(config_bytes_pattern, "Current configuration : CONFIG_BYTES", response)
        return response

    def _replace_timestamps(response):
        datetime_pattern = re.compile(
            r"\d+:\d+:\d+\d+\s+[a-z]{3}\s+(mon|tue|wed|thu|fri|sat|sun)\s+(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+\d+\s+\d+",
            flags=re.M | re.I,
        )
        response = re.sub(datetime_pattern, "TIME_STAMP_REPLACED", response)
        return response

    def _replace_configured_by(response):
        configured_by_pattern = re.compile(
            r"^! Last configuration change at TIME_STAMP_REPLACED by (\w+)$", flags=re.M | re.I
        )
        response = re.sub(
            configured_by_pattern, "! Last configuration change at TIME_STAMP_REPLACED", response
        )
        return response

    def _replace_crypto_strings(response):
        crypto_pattern = re.compile(
            r"^\s+certificate self-signed.*$\s(^\s{2}(\w+\s){1,8})+\s+quit$", flags=re.M | re.I
        )
        response = re.sub(crypto_pattern, "CRYPTO_REPLACED", response)
        return response

    def _replace_hashed_passwords(response):
        crypto_pattern = re.compile(r"^enable secret 5 (.*$)", flags=re.M | re.I)
        response = re.sub(crypto_pattern, "enable secret 5 HASHED_PASSWORD", response)
        return response

    response = _replace_config_bytes(response)
    response = _replace_timestamps(response)
    response = _replace_configured_by(response)
    response = _replace_crypto_strings(response)
    response = _replace_hashed_passwords(response)
    return response


def cisco_iosxr_clean_response(response):
    def _replace_timestamps(response):
        datetime_pattern = re.compile(
            r"(mon|tue|wed|thu|fri|sat|sun)\s+(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+\d+\s+\d+:\d+:\d+((\.\d+\s\w+)|\s\d+)",
            flags=re.M | re.I,
        )
        response = re.sub(datetime_pattern, "TIME_STAMP_REPLACED", response)
        return response

    def _replace_configured_by(response):
        configured_by_pattern = re.compile(
            r"^!! Last configuration change at TIME_STAMP_REPLACED by (\w+)$", flags=re.M | re.I
        )
        response = re.sub(
            configured_by_pattern, "!! Last configuration change at TIME_STAMP_REPLACED", response
        )
        return response

    def _replace_crypto_strings(response):
        crypto_pattern = re.compile(r"^\ssecret\s5\s[\w$\.\/]+$", flags=re.M | re.I)
        response = re.sub(crypto_pattern, "CRYPTO_REPLACED", response)
        return response

    def _replace_commit_in_progress(response):
        commit_in_progress_pattern = re.compile(r"System configuration.*", flags=re.M | re.I)
        response = re.sub(commit_in_progress_pattern, "", response)
        return response

    response = _replace_timestamps(response)
    response = _replace_configured_by(response)
    response = _replace_crypto_strings(response)
    response = _replace_commit_in_progress(response)
    return response


def cisco_nxos_clean_response(response):
    def _replace_timestamps(response):
        datetime_pattern = re.compile(
            r"(mon|tue|wed|thu|fri|sat|sun)\s+(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+\d+\s+\d+:\d+:\d+\s\d+",
            flags=re.M | re.I,
        )
        response = re.sub(datetime_pattern, "TIME_STAMP_REPLACED", response)
        return response

    def _replace_crypto_strings(response):
        crypto_pattern = re.compile(r"^(.*?\s(?:5|md5)\s)[\w$\.\/]+.*$", flags=re.M | re.I)
        response = re.sub(crypto_pattern, "CRYPTO_REPLACED", response)
        return response

    response = _replace_timestamps(response)
    response = _replace_crypto_strings(response)
    return response


def arista_eos_clean_response(response):
    def _replace_timestamps(response):
        datetime_pattern = re.compile(
            r"(mon|tue|wed|thu|fri|sat|sun)\s+(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+\d+\s+\d+:\d+:\d+\s+\d+$",
            flags=re.M | re.I,
        )
        response = re.sub(datetime_pattern, "TIME_STAMP_REPLACED", response)
        return response

    def _replace_crypto_strings(response):
        crypto_pattern = re.compile(r"secret\ssha512\s[\w$\.\/]+$", flags=re.M | re.I)
        response = re.sub(crypto_pattern, "CRYPTO_REPLACED", response)
        return response

    response = _replace_timestamps(response)
    response = _replace_crypto_strings(response)
    return response


def juniper_junos_clean_response(response):
    def _replace_timestamps(response):
        datetime_pattern = re.compile(
            r"^## Last commit: \d+-\d+-\d+\s\d+:\d+:\d+\s\w+.*$", flags=re.M | re.I
        )
        response = re.sub(datetime_pattern, "TIME_STAMP_REPLACED", response)
        return response

    def _replace_crypto_strings(response):
        crypto_pattern = re.compile(
            r'^\s+encrypted-password\s"[\w$\.\/]+";\s.*$', flags=re.M | re.I
        )
        response = re.sub(crypto_pattern, "CRYPTO_REPLACED", response)
        return response

    response = _replace_timestamps(response)
    response = _replace_crypto_strings(response)
    return response



"""mock_devices.base"""
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

import asyncssh


class BaseSSHServerSession(ABC, asyncssh.SSHServerSession):  # type: ignore
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        SSH Server Session class

        Inherits from asyncssh and provides some extra context/setup for the mock network devices

        Args:
            args: positional args to pass to base class
            kwargs: keyword args to pass to base class

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(*args, **kwargs)

        self._chan: asyncssh.editor.SSHLineEditorChannel

        self._hide_input = False
        self._interacting = False

        self.terminal_length = 80
        self.terminal_width = 80

        self.priv_levels: Dict[str, str] = {}
        self.priv_level = ""
        self.invalid_input = ""
        self.paging_prompt = ""

        self.command_mapper: Dict[Callable[..., Any], List[str]] = {
            self.handle_open: [],
            self.handle_close: [],
            self.handle_priv: [],
            self.handle_show: [],
            self.handle_interactive: [],
        }

        self.show_command_output = ""

        self._redirect_to_handler: Optional[Callable[..., Any]] = None

    def connection_made(self, chan: asyncssh.editor.SSHLineEditorChannel) -> None:
        """
        SSH Connection made!

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self._chan = chan

    def shell_requested(self) -> bool:
        """
        Handle shell requested; always return True

        Args:
            N/A

        Returns:
            bool: always True!

        Raises:
            N/A

        """
        return True

    def session_started(self) -> None:
        """
        SSH session started

        Initial SSH session started; present user with the priv level set at priv_level

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self._chan.write(self.priv_levels.get(self.priv_level))

    @abstractmethod
    def handle_open(self, channel_input: str) -> None:
        """
        Handle "open" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def handle_close(self, channel_input: str) -> None:
        """
        Handle "close" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def handle_priv(self, channel_input: str) -> None:
        """
        Handle "privilege" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def handle_invalid(self, channel_input: str) -> None:
        """
        Handle "invalid" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = channel_input  # maybe care about this in the future or for some platforms
        self._chan.write(self.invalid_input)
        self.repaint_prompt()

    def handle_show(self, channel_input: str) -> None:
        """
        Handle "show" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = channel_input  # we dont care about this, we only support one show command anyway :)
        if self.terminal_length:
            channel_output = "\n".join(
                self.show_command_output.splitlines()[: self.terminal_length]
            )
            channel_output += self.paging_prompt
        else:
            channel_output = self.show_command_output
            channel_output += f"\n{self.priv_levels.get(self.priv_level)}"
        self._chan.write(channel_output)

    @abstractmethod
    def handle_interactive(self, channel_input: str) -> None:
        """
        Handle "interactive" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def repaint_prompt(self) -> None:
        """
        Paint the prompt to the ssh channel

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self._chan.write(self.priv_levels.get(self.priv_level))

    @abstractmethod
    def data_received(self, data: str, datatype: None) -> None:
        """
        Handle data received on ssh channel

        Args:
            data: string of data sent to channel
            datatype: dunno! in base class though :)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def eof_received(self) -> None:
        """
        Handle eof

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self._chan.exit(0)

    def break_received(self, msec: float) -> None:
        """
        Handle break

        Args:
            msec: dunno, but in base class implementation :)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.eof_received()


class BaseServer(asyncssh.SSHServer):  # type: ignore
    def session_requested(self) -> asyncssh.SSHServerSession:
        """
        Session requested; return ServerSession object

        `ServerSession` set in `run` to be the appropriate SSHServerSession type for a given
        platform, i.e. `IOSXESSHServerSession`

        Args:
            N/A

        Returns:
            asyncssh.SSHServerSession: SSHServerSession

        Raises:
            N/A

        """
        return self.ServerSession()

    def begin_auth(self, username: str) -> bool:
        """
        Begin auth; always returns True

        Args:
            username: username for auth

        Returns:
            bool: always True

        Raises:
            N/A

        """
        return True

    def password_auth_supported(self) -> bool:
        """
        Password auth supported; always return True

        Args:
            N/A

        Returns:
            bool: always True

        Raises:
            N/A

        """
        return True

    def public_key_auth_supported(self) -> bool:
        """
        Public key auth supported; always return True

        Args:
            N/A

        Returns:
            bool: always True

        Raises:
            N/A

        """
        return True

    def validate_password(self, username: str, password: str) -> bool:
        """
        Validate provided username/password

        Args:
            username: username to check for auth
            password: password to check for auth

        Returns:
            bool: True if user/password is correct (scrapli/scrapli)

        Raises:
            N/A

        """
        if username == password == "scrapli":
            return True
        return False

    def validate_public_key(
        self, username: str, key: asyncssh.rsa._RSAKey  # pylint: disable=W0212
    ) -> bool:
        """
        Validate provided public key

        Args:
            username: username to check for auth
            key: asyncssh RSAKey to check for auth

        Returns:
            bool: True if ssh key is correct

        Raises:
            N/A

        """
        if (
            username == "scrapli"
            and key.get_fingerprint() == "SHA256:rb1CVtQCkWBAzm1AxV7xR7BLBawUwFUlUVFVu+QYQBM"
        ):
            return True
        return False



"""scrapli.factory"""
import importlib
from copy import deepcopy
from logging import getLogger
from typing import Any, Dict, Optional, Tuple, Type, Union

from scrapli.driver import AsyncGenericDriver, AsyncNetworkDriver, GenericDriver, NetworkDriver
from scrapli.driver.base_driver import ASYNCIO_TRANSPORTS
from scrapli.driver.core import (
    AsyncEOSDriver,
    AsyncIOSXEDriver,
    AsyncIOSXRDriver,
    AsyncJunosDriver,
    AsyncNXOSDriver,
    EOSDriver,
    IOSXEDriver,
    IOSXRDriver,
    JunosDriver,
    NXOSDriver,
)
from scrapli.exceptions import ScrapliException

LOG = getLogger("scrapli.factory")

ASYNC_CORE_PLATFORM_MAP = {
    "arista_eos": AsyncEOSDriver,
    "cisco_iosxe": AsyncIOSXEDriver,
    "cisco_iosxr": AsyncIOSXRDriver,
    "cisco_nxos": AsyncNXOSDriver,
    "juniper_junos": AsyncJunosDriver,
}
SYNC_CORE_PLATFORM_MAP = {
    "arista_eos": EOSDriver,
    "cisco_iosxe": IOSXEDriver,
    "cisco_iosxr": IOSXRDriver,
    "cisco_nxos": NXOSDriver,
    "juniper_junos": JunosDriver,
}
ASYNC_DRIVER_MAP = {"network": AsyncNetworkDriver, "generic": AsyncGenericDriver}
SYNC_DRIVER_MAP = {"network": NetworkDriver, "generic": GenericDriver}


def _get_community_platform_details(community_platform_name: str) -> Dict[str, Any]:
    """
    Parent get driver method

    Args:
        community_platform_name: name of community

    Returns:
        platform_details: dict of details about community platform from scrapli_community library

    Raises:
        ModuleNotFoundError: if scrapli_community is not importable
        ScrapliException: if community platform is missing "SCRAPLI_PLATFORM" attribute
        ScrapliException: for any unknown exception during community platform import

    """
    try:
        # replace any underscores in platform name with "."; should support any future platforms
        # that dont have "child" os types -- i.e. just "cisco" instead of "cisco_iosxe"
        scrapli_community_platform = importlib.import_module(
            name=f"scrapli_community.{community_platform_name.replace('_', '.')}"
        )
    except ModuleNotFoundError as exc:
        err = f"Module '{exc.name}' not found!"
        msg = f"***** {err} {'*' * (80 - len(err))}"
        fix = (
            "To resolve this issue, ensure you have the scrapli community package installed."
            " You can install this with pip: `pip install scrapli_community`."
        )
        warning = "\n" + msg + "\n" + fix + "\n" + msg
        raise ModuleNotFoundError(warning) from exc
    except Exception as exc:
        msg = "Unknown error occurred"
        raise ScrapliException(msg) from exc

    platform_details_original = getattr(scrapli_community_platform, "SCRAPLI_PLATFORM", {})
    if not platform_details_original:
        msg = "Community platform missing required attribute `SCRAPLI_PLATFORM`"
        raise ScrapliException(msg)
    platform_details: Dict[str, Any] = deepcopy(platform_details_original)
    return platform_details


def _get_community_driver(
    community_platform_name: str, variant: Optional[str], _async: bool = False
) -> Tuple[
    Union[
        Type[AsyncNetworkDriver], Type[AsyncGenericDriver], Type[NetworkDriver], Type[GenericDriver]
    ],
    Dict[str, Any],
]:
    """
    Parent get driver method

    Args:
        community_platform_name: name of community
        variant: optional name of variant of community platform
        _async: True/False this is for an asyncio transport driver

    Returns:
        NetworkDriver: final driver class

    Raises:
        ScrapliException: if scrapli_community platform has an invalid value for "driver_type"

    """
    platform_details = _get_community_platform_details(
        community_platform_name=community_platform_name
    )

    driver_type = platform_details["driver_type"]
    if _async is False:
        final_driver = SYNC_DRIVER_MAP.get(driver_type, None)
    else:
        final_driver = ASYNC_DRIVER_MAP.get(driver_type, None)
    if not final_driver:
        raise ScrapliException("Invalid driver type provided in community platform data")

    platform_kwargs = platform_details["defaults"]

    if variant:
        variant_kwargs = platform_details["variants"][variant]
        final_platform_kwargs = {**platform_kwargs, **variant_kwargs}
    else:
        final_platform_kwargs = platform_kwargs

    if not _async:
        # remove unnecessary asyncio things
        final_platform_kwargs.pop("async_on_open")
        final_platform_kwargs.pop("async_on_close")
        # rename sync_on_(open|close) keys to just "on_open"/"on_close"
        final_platform_kwargs["on_open"] = final_platform_kwargs.pop("sync_on_open")
        final_platform_kwargs["on_close"] = final_platform_kwargs.pop("sync_on_close")
    else:
        # remove unnecessary sync things
        final_platform_kwargs.pop("sync_on_open")
        final_platform_kwargs.pop("sync_on_close")
        # rename sync_on_(open|close) keys to just "on_open"/"on_close"
        final_platform_kwargs["on_open"] = final_platform_kwargs.pop("async_on_open")
        final_platform_kwargs["on_close"] = final_platform_kwargs.pop("async_on_close")

    return final_driver, final_platform_kwargs


def _get_driver(
    platform: str, variant: Optional[str], _async: bool = False
) -> Tuple[Union[Type[NetworkDriver], Type[GenericDriver]], Dict[str, Any]]:
    """
    Parent get driver method

    Args:
        platform: name of target platform; i.e. `cisco_iosxe`, `arista_eos`, etc.
        variant: name of the target paltform variant
        _async: True/False this is for an asyncio transport driver

    Returns:
        NetworkDriver: final driver class; generally NetworkDriver, but for some community platforms
            could be GenericDriver

    Raises:
        N/A

    """
    additional_kwargs: Dict[str, Any] = {}

    if platform in SYNC_CORE_PLATFORM_MAP:
        if _async is False:
            final_driver = SYNC_CORE_PLATFORM_MAP[platform]
        else:
            final_driver = ASYNC_CORE_PLATFORM_MAP[platform]
        msg = f"Driver `{final_driver}` selected from scrapli core drivers"
    else:
        final_driver, additional_kwargs = _get_community_driver(
            community_platform_name=platform, variant=variant, _async=_async
        )
        msg = (
            f"Driver `{final_driver}` selected from scrapli community platforms, with the following"
            f" platform arguments: `{additional_kwargs}`"
        )

    LOG.info(msg)
    return final_driver, additional_kwargs


class Scrapli(NetworkDriver):
    def __new__(
        cls, platform: str, variant: Optional[str] = None, **kwargs: Dict[Any, Any]
    ) -> "Scrapli":
        """
        Scrapli Factory method for synchronous drivers

        Args:
            cls: class object
            platform: name of target platform; i.e. `cisco_iosxe`, `arista_eos`, etc.
            variant: optional name of variant of community platform
            **kwargs: keyword arguments to pass to selected driver class

        Returns:
            final_driver: synchronous driver class for provided driver

        Raises:
            ScrapliException: if provided transport is asyncio
            ScrapliException: if `platform` not in keyword arguments

        """
        LOG.debug("Scrapli factory initialized")

        if kwargs.get("transport", "system") in ASYNCIO_TRANSPORTS:
            raise ScrapliException("Use `AsyncScrapli` if using an async transport!")

        if not isinstance(platform, str):
            raise ScrapliException(f"Argument `platform` must be `str` got `{type(platform)}`")

        final_driver, additional_kwargs = _get_driver(
            platform=platform, variant=variant, _async=False
        )

        # at this point will need to merge the additional kwargs in (for community drivers),
        # ensure that kwargs passed by user supersede the ones coming from community platform
        if additional_kwargs:
            final_kwargs = {**additional_kwargs, **kwargs}
        else:
            final_kwargs = kwargs

        # mypy was displeased about NetworkDriver not being callable, fix later probably :)
        final_conn: "Scrapli" = final_driver(**final_kwargs)  # type: ignore
        return final_conn


class AsyncScrapli(AsyncNetworkDriver):
    def __new__(
        cls, platform: str, variant: Optional[str] = None, **kwargs: Dict[Any, Any]
    ) -> "AsyncScrapli":
        """
        Scrapli Factory method for asynchronous drivers

        Args:
            cls: class object
            platform: name of target platform; i.e. `cisco_iosxe`, `arista_eos`, etc.
            variant: optional name of variant of community platform
            **kwargs: keyword arguments to pass to selected driver class

        Returns:
            final_driver: synchronous driver class for provided driver

        Raises:
            ScrapliException: if provided transport is not asyncio
            ScrapliException: if `platform` not in keyword arguments

        """
        LOG.debug("Scrapli factory initialized")

        if kwargs.get("transport", "system") not in ASYNCIO_TRANSPORTS:
            raise ScrapliException("Use `Scrapli` if using a synchronous transport!")

        if not isinstance(platform, str):
            raise ScrapliException(f"Argument `platform` must be `str` got `{type(platform)}`")

        final_driver, additional_kwargs = _get_driver(
            platform=platform, variant=variant, _async=True
        )

        # at this point will need to merge the additional kwargs in (for community drivers),
        # ensure that kwargs passed by user supersede the ones coming from community platform
        if additional_kwargs:
            final_kwargs = {**additional_kwargs, **kwargs}
        else:
            final_kwargs = kwargs

        # mypy was displeased about NetworkDriver not being callable, fix later probably :)
        final_conn: "AsyncScrapli" = final_driver(**final_kwargs)  # type: ignore
        return final_conn



import re

import pytest

from scrapli.driver.core.cisco_nxos.driver import PRIVS


@pytest.mark.parametrize(
    "priv_pattern",
    [
        ("exec", "switch>"),
        ("privilege_exec", "switch# "),
        ("configuration", "switch(config)# "),
        ("privilege_exec", "swi_tch# "),
    ],
    ids=[
        "exec",
        "privilege_exec",
        "configuration",
        "underscore_privilege_exec",
    ],
)
def test_prompt_patterns(priv_pattern):
    priv_level_name = priv_pattern[0]
    prompt = priv_pattern[1]
    prompt_pattern = PRIVS.get(priv_level_name).pattern
    match = re.search(pattern=prompt_pattern, string=prompt, flags=re.M | re.I)
    assert match



import re

import pytest

from scrapli.driver.core.juniper_junos.driver import PRIVS


@pytest.mark.parametrize(
    "priv_pattern",
    [("exec", "vrnetlab> "), ("configuration", "vrnetlab# "), ("exec", "vrn_etlab> ")],
    ids=["exec", "configuration", "exec_underscore"],
)
def test_prompt_patterns(priv_pattern):
    priv_level_name = priv_pattern[0]
    prompt = priv_pattern[1]
    prompt_pattern = PRIVS.get(priv_level_name).pattern
    match = re.search(pattern=prompt_pattern, string=prompt, flags=re.M | re.I)
    assert match



def test_non_standard_default_desired_privilege_level(iosxe_conn):
    # purpose of this test is to ensure that when a user sets a non-standard default desired priv
    # level, that there is nothing in genericdriver/networkdriver that will prevent that from
    # actually being set as the default desired priv level
    iosxe_conn.close()
    iosxe_conn.default_desired_privilege_level = "configuration"
    iosxe_conn.open()
    current_prompt = iosxe_conn.get_prompt()
    assert current_prompt == "csr1000v(config)#"
    iosxe_conn.close()



from typing import Optional

from scrapli.transport import Transport


class MockTransport(Transport):
    def __init__(self, host):
        super().__init__()
        self.host = host

    def open(self) -> None:
        return

    def close(self) -> None:
        return

    def isalive(self) -> bool:
        return True

    def read(self) -> bytes:
        return b""

    def write(self, channel_input: str) -> None:
        return

    def flush(self) -> None:
        return

    def set_timeout(self, timeout: Optional[int] = None) -> None:
        return

    def _session_keepalive(self) -> None:
        return

    def _keepalive_standard(self) -> None:
        return


def test_transport_abc_init():
    transport = MockTransport("racecar")
    assert transport.host == "racecar"



"""scrapli.driver.base_network_driver"""
import re
import warnings
from collections import UserList
from datetime import datetime
from enum import Enum
from functools import lru_cache
from logging import Logger
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from scrapli.exceptions import UnknownPrivLevel
from scrapli.helper import resolve_file
from scrapli.response import Response

if TYPE_CHECKING:
    ScrapliMultiResponse = UserList[Response]  # pylint:  disable=E1136; # pragma:  no cover
else:
    ScrapliMultiResponse = UserList


class PrivilegeLevel:
    __slots__ = (
        "pattern",
        "name",
        "previous_priv",
        "deescalate",
        "escalate",
        "escalate_auth",
        "escalate_prompt",
    )

    def __init__(
        self,
        pattern: str,
        name: str,
        previous_priv: str,
        deescalate: str,
        escalate: str,
        escalate_auth: bool,
        escalate_prompt: str,
    ):
        """
        PrivilegeLevel Object

        Args:
            pattern: regex pattern to use to identify this privilege level by the prompt
            name: friendly name of this privilege level
            previous_priv: name of the lower/previous privilege level
            deescalate: how to deescalate *from* this privilege level (to the lower/previous priv)
            escalate: how to escalate *to* this privilege level (from the lower/previous priv)
            escalate_auth: True/False escalation requires authentication
            escalate_prompt: prompt pattern to search for during escalation if escalate auth is True

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.pattern = pattern
        self.name = name
        self.previous_priv = previous_priv
        self.deescalate = deescalate
        self.escalate = escalate
        self.escalate_auth = escalate_auth
        self.escalate_prompt = escalate_prompt


DUMMY_PRIV_LEVEL = PrivilegeLevel("", "DUMMY", "", "", "", False, "")
PRIVS: Dict[str, PrivilegeLevel] = {}


class PrivilegeAction(Enum):
    NO_ACTION = "no action"
    ESCALATE = "escalate"
    DEESCALATE = "deescalate"


class NetworkDriverBase:
    # NetworkDriverBase Mixin vars for typing/linting purposes
    logger: Logger
    comms_prompt_pattern: str
    _current_priv_level = DUMMY_PRIV_LEVEL
    _priv_map: Dict[str, List[str]]
    failed_when_contains: Optional[List[str]]
    privilege_levels: Dict[str, PrivilegeLevel]
    auth_secondary: str
    textfsm_platform: str
    genie_platform: str

    @staticmethod
    def _check_kwargs_comms_prompt_pattern(kwargs: Any) -> None:
        """
        Warn users if passing comms prompt pattern while using NetworkDriver derived classes

        Args:
            kwargs: kwargs being passed into NetworkDriver classes

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if "comms_prompt_pattern" in kwargs:
            err = "`comms_prompt_pattern` found in kwargs!"
            msg = f"***** {err} {'*' * (80 - len(err))}"
            fix = (
                "`comms_prompt_pattern` is ignored (dropped) when using network drivers. If you "
                "wish to modify the patterns for any network driver sub-classes, please do so by "
                "modifying or providing your own `privilege_levels`."
            )
            warning = "\n" + msg + "\n" + fix + "\n" + msg
            warnings.warn(warning)
            kwargs.pop("comms_prompt_pattern")

    def _generate_comms_prompt_pattern(self) -> None:
        """
        Generate the `comms_prompt_pattern_all` from the currently assigned privilege levels

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.comms_prompt_pattern = r"|".join(
            rf"({priv_level_data.pattern})" for priv_level_data in self.privilege_levels.values()
        )

    def _build_priv_map(self) -> None:
        """
        Build a "map" of privilege levels

        `_priv_map` is a "map" of all privilege levels mapped out to the lowest available priv. This
        map is used for determining how to escalate/deescalate.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        for priv_level in self.privilege_levels:
            self._priv_map[priv_level] = [priv_level]
            while True:
                previous_priv = self.privilege_levels[self._priv_map[priv_level][0]].previous_priv
                if not previous_priv:
                    break
                self._priv_map[priv_level].insert(0, previous_priv)

    def update_privilege_levels(self, update_channel: bool = True) -> None:
        """
        Re-generate the privilege map, and update the comms prompt pattern

        Args:
            update_channel: True/False update the channel pattern too -- likely only ever set to
                False for class initialization before channel is opened

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self._build_priv_map()
        self._generate_comms_prompt_pattern()
        # clear the lru cache as patterns may have been updated
        self._determine_current_priv.cache_clear()
        if update_channel is True:
            self.channel.comms_prompt_pattern = (  # type: ignore  # pylint: disable=E1101
                self.comms_prompt_pattern
            )

    @lru_cache()
    def _determine_current_priv(self, current_prompt: str) -> List[str]:
        """
        Determine current privilege level from prompt string

        Args:
            current_prompt: string of current prompt

        Returns:
            matching_priv_levels: list of string names of matching privilege levels

        Raises:
            UnknownPrivLevel: if privilege level cannot be determined

        """
        matching_priv_levels = []
        for priv_level in self.privilege_levels.values():
            search_result = re.search(
                pattern=priv_level.pattern, string=current_prompt, flags=re.M | re.I
            )
            if not search_result:
                continue
            matching_priv_levels.append(priv_level.name)
            self.logger.debug(f"Current privilege level could be `{priv_level.name}`")
        if not matching_priv_levels:
            raise UnknownPrivLevel(
                f"Could not determine privilege level from provided prompt: `{current_prompt}`"
            )
        self.logger.debug(f"Determined current privilege level is one of `{matching_priv_levels}`")
        return matching_priv_levels

    def _get_privilege_level_name(self, requested_priv: str) -> str:
        """
        Get privilege level name if provided privilege is valid

        Args:
            requested_priv: string name of desired privilege level

        Returns:
            str: name of the privilege level requested

        Raises:
            UnknownPrivLevel: if attempting to acquire an unknown priv

        """
        desired_privilege_level = self.privilege_levels.get(requested_priv, None)
        if desired_privilege_level is None:
            raise UnknownPrivLevel(
                f"Requested privilege level `{requested_priv}` not a valid privilege level of "
                f"`{self.__class__.__name__}`"
            )
        resolved_privilege_level = desired_privilege_level.name
        return resolved_privilege_level

    def _update_response(self, response: Response) -> None:
        """
        Update response with network driver specific data

        This happens here as the underlying channel provides a response object but is unaware of any
        of the network/platform specific attributes that may need to get updated

        Args:
            response: response to update

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        response.textfsm_platform = self.textfsm_platform
        response.genie_platform = self.genie_platform

    def register_configuration_session(self, session_name: str) -> None:
        """
        If applicable, register a configuration session as a valid privilege level

        Args:
            session_name: name of config session to register

        Returns:
            N/A:  # noqa: DAR202

        Raises:
            NotImplementedError: unless overridden by concrete class

        """
        _ = session_name
        raise NotImplementedError(
            f"Configuration sessions not supported for `{self.__class__.__name__}`"
        )

    def _pre_escalate(self, escalate_priv: PrivilegeLevel) -> None:
        """
        Handle pre "_escalate" tasks for consistency between sync/async versions

        Args:
            escalate_priv: privilege level to escalate to

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if escalate_priv.escalate_auth is True and not self.auth_secondary:
            err = (
                "Privilege escalation generally requires an `auth_secondary` password, "
                "but none is set!"
            )
            msg = f"***** {err} {'*' * (80 - len(err))}"
            fix = (
                "scrapli will try to escalate privilege without entering a password but may "
                "fail.\nSet an `auth_secondary` password if your device requires a password to "
                "increase privilege, otherwise ignore this message."
            )
            warning = "\n" + msg + "\n" + fix + "\n" + msg
            warnings.warn(warning)

    def _pre_acquire_priv(self, desired_priv: str) -> Tuple[str, List[str]]:
        """
        Handle pre "acquire_priv" tasks for consistency between sync/async versions

        Args:
            desired_priv: string name of desired privilege level

        Returns:
            resolved_priv: privilege level class resolved via given privilege name
            map_to_desired_priv: map from current priv to desired priv (list of strings of privs to
                iterate through to get to target priv)

        Raises:
            N/A

        """
        self.logger.info(f"Attempting to acquire `{desired_priv}` privilege level")
        resolved_priv = self._get_privilege_level_name(requested_priv=desired_priv)
        map_to_desired_priv = self._priv_map[resolved_priv]
        return resolved_priv, map_to_desired_priv

    def _process_acquire_priv(
        self,
        resolved_priv: str,
        map_to_desired_priv: List[str],
        current_prompt: str,
    ) -> Tuple[PrivilegeAction, PrivilegeLevel]:
        """
        Handle non channel "acquire_priv" tasks for consistency between sync/async versions

        Args:
            resolved_priv: string name of desired privilege level
            map_to_desired_priv: list of string names of privilege levels to acquire to get to
                desired privilege level
            current_prompt: string of the current prompt

        Returns:
            PrivilegeAction: enum set to appropriate value for no action, escalate or deescalate
            PrivilegeLevel: privilege level object to pass to either escalate or deescalate method

        Raises:
            N/A

        """
        # if we are already at the desired priv, we don't need to do any thing
        current_priv_patterns = self._determine_current_priv(current_prompt=current_prompt)

        # if multiple patterns match pick the zeroith... the only time patterns should be
        # identical is if we have privilege levels like "configuration" or
        # "configuration_exclusive" that have identical prompts (ex: IOSXRDriver)
        current_priv = self.privilege_levels[current_priv_patterns[0]]

        if resolved_priv in current_priv_patterns:
            self.logger.info(f"Acquired requested privilege level `{resolved_priv}`")
            self._current_priv_level = self.privilege_levels[resolved_priv]
            return PrivilegeAction.NO_ACTION, current_priv

        map_to_current_priv = self._priv_map[current_priv.name]
        priv_map = (
            map_to_current_priv
            if map_to_current_priv > map_to_desired_priv
            else map_to_desired_priv
        )

        desired_priv_index = priv_map.index(resolved_priv)
        try:
            current_priv_index = priv_map.index(current_priv.name)
        except ValueError:
            # if the current priv is not in the map for the desired priv; set the current index
            # to the "top" (end) of the priv map and work our way back down
            current_priv_index = len(priv_map)

        if current_priv_index > desired_priv_index:
            deescalate_priv = priv_map[current_priv_index - 1]
            self.logger.info(
                f"Attempting to deescalate from {current_priv.name} to {deescalate_priv}"
            )
            return PrivilegeAction.DEESCALATE, current_priv

        escalate_priv = self.privilege_levels[priv_map[current_priv_index + 1]]
        self.logger.info(f"Attempting to escalate from {current_priv.name} to {escalate_priv.name}")
        return PrivilegeAction.ESCALATE, escalate_priv

    @staticmethod
    def _pre_send_config(config: str) -> List[str]:
        """
        Handle pre "send_config" tasks for consistency between sync/async versions

        Args:
            config: string configuration to send to the device, supports sending multi-line strings

        Returns:
            configs: list of config lines from provided "config" input

        Raises:
            TypeError: if anything but a string is provided for `file`

        """
        if not isinstance(config, str):
            raise TypeError(
                f"`send_config` expects a single string, got {type(config)}, "
                "to send a list of configs use the `send_configs` method instead."
            )

        # in order to handle multi-line strings, we split lines
        split_config = config.splitlines()

        return split_config

    def _post_send_config(
        self,
        config: str,
        multi_response: ScrapliMultiResponse,
    ) -> Response:
        """
        Handle post "send_config" tasks for consistency between sync/async versions

        Args:
            config: string configuration to send to the device, supports sending multi-line strings
            multi_response: multi_response object send_config got from calling self.send_configs;
                we need this to parse out the multi_response back into a single Response object

        Returns:
            Response: Unified response object

        Raises:
            N/A

        """
        # capture failed_when_contains and host from zeroith multi_response element (there should
        #  always be at least a zeroith element here!); getting host just lets us keep the mixin
        #  class a little cleaner without having to deal with sync vs async transport classes from
        #  a typing perspective
        failed_when_contains = multi_response[0].failed_when_contains
        host = multi_response[0].host

        # create a new unified response object
        response = Response(
            host=host,
            channel_input=config,
            failed_when_contains=failed_when_contains,
        )
        response.start_time = multi_response[0].start_time
        response.elapsed_time = (datetime.now() - response.start_time).total_seconds()

        # join all the results together into a single final result
        response.result = "\n".join(response.result for response in multi_response)
        response.failed = False

        if any(response.failed for response in multi_response):
            response.failed = True
        self._update_response(response=response)

        return response

    def _pre_send_configs(
        self,
        configs: List[str],
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        privilege_level: str = "",
    ) -> Tuple[str, Optional[Union[str, List[str]]]]:
        """
        Handle pre "send_configs" tasks for consistency between sync/async versions

        Args:
            configs: list of strings to send to device in config mode
            failed_when_contains: string or list of strings indicating failure if found in response
            privilege_level: name of configuration privilege level/type to acquire; this is platform
                dependent, so check the device driver for specifics. Examples of privilege_name
                would be "configuration_exclusive" for IOSXRDriver, or "configuration_private" for
                JunosDriver. You can also pass in a name of a configuration session such as
                "my-config-session" if you have registered a session using the
                "register_config_session" method of the EOSDriver or NXOSDriver.

        Returns:
            commands: list of commands read from file

        Raises:
            TypeError: if configs is anything but a list

        """
        if not isinstance(configs, list):
            raise TypeError(
                f"`send_configs` expects a list of strings, got {type(configs)}, "
                "to send a single configuration line/string use the `send_config` method instead."
            )

        if failed_when_contains is None:
            failed_when_contains = self.failed_when_contains

        if privilege_level:
            resolved_privilege_level = self._get_privilege_level_name(
                requested_priv=privilege_level
            )
        else:
            resolved_privilege_level = "configuration"

        return resolved_privilege_level, failed_when_contains

    def _post_send_configs(self, responses: ScrapliMultiResponse) -> ScrapliMultiResponse:
        """
        Handle post "send_configs" tasks for consistency between sync/async versions

        Args:
            responses: multi_response object to update

        Returns:
            ScrapliMultiResponse: Unified response object

        Raises:
            N/A

        """
        for response in responses:
            self._update_response(response=response)

        return responses

    @staticmethod
    def _pre_send_configs_from_file(file: str) -> List[str]:
        """
        Handle pre "send_configs_from_file" tasks for consistency between sync/async versions

        Args:
            file: string path to file

        Returns:
            commands: list of commands read from file

        Raises:
            TypeError: if anything but a string is provided for `file`

        """
        if not isinstance(file, str):
            raise TypeError(
                f"`send_configs_from_file` expects a string path to a file, got {type(file)}"
            )
        resolved_file = resolve_file(file)

        with open(resolved_file, "r") as f:
            configs = f.read().splitlines()

        return configs



import pytest

from scrapli.driver.core import AsyncIOSXEDriver
from scrapli.exceptions import TransportPluginError

from ...test_data.devices import DEVICES


@pytest.mark.asyncio
async def test_context_manager():
    device = DEVICES["mock_cisco_iosxe"].copy()
    async with AsyncIOSXEDriver(transport="asyncssh", **device) as conn:
        assert conn.isalive() is True
    assert conn.isalive() is False


@pytest.mark.asyncio
async def test_transport_bool(async_cisco_iosxe_conn):
    await async_cisco_iosxe_conn.open()
    assert bool(async_cisco_iosxe_conn.transport) is True
    await async_cisco_iosxe_conn.close()
    assert bool(async_cisco_iosxe_conn.transport) is False


@pytest.mark.asyncio
async def test_non_async_transport():
    device = DEVICES["mock_cisco_iosxe"].copy()
    with pytest.raises(TransportPluginError) as exc:
        AsyncIOSXEDriver(transport="system", **device)
    assert (
        str(exc.value)
        == "Attempting to use transport type system with an asyncio driver, must use one of ['asyncssh'] transports"
    )


def test_timeout_ops_property(async_cisco_iosxe_conn):
    # note that there is "sync" and "async" version of this test as the property setter/getter lives
    # in the sync and async drivers -- didnt want to deal w/ having the property in the base driver
    # and having to deal w/ _get_timeout_ops/_set_timeout_ops methods to get overridden by the sync
    # and async drivers
    assert async_cisco_iosxe_conn.timeout_ops == 30
    assert async_cisco_iosxe_conn.channel.timeout_ops == 30
    async_cisco_iosxe_conn.timeout_ops = 5
    assert async_cisco_iosxe_conn.timeout_ops == 5
    assert async_cisco_iosxe_conn.channel.timeout_ops == 5



"""mock_devices.run"""
import asyncio
from typing import List, Optional

import asyncssh

from .cisco_iosxe import IOSXEServer, IOSXESSHServerSession
from .cisco_nxos import NXOSServer, NXOSSSHServerSession
from .juniper_junos import JunosServer, JunoSSHServerSession

SERVERS = {
    "cisco_iosxe": {"server_session": IOSXESSHServerSession, "server": IOSXEServer, "port": 2211},
    "cisco_nxos": {"server_session": NXOSSSHServerSession, "server": NXOSServer, "port": 2212},
    "juniper_junos": {"server_session": JunoSSHServerSession, "server": JunosServer, "port": 2215},
}


async def start_server(platform: str, server_key: str) -> None:
    """
    Coroutine to start ssh server

    Args:
        platform: name of server type to start
        server_key: string path to key to use for ssh server(s)

    Returns:
        None

    Raises:
        N/A

    """
    server = SERVERS[platform]["server"]
    server.ServerSession = SERVERS[platform]["server_session"]
    port = SERVERS[platform]["port"]
    await asyncssh.create_server(server, "localhost", port, server_host_keys=[server_key])


async def main(server_key: str, servers: List[str]) -> None:
    """
    Async main

    Args:
        server_key: string path to key to use for ssh server(s)
        servers: list of servers to start

    Returns:
        None

    Raises:
        N/A

    """
    await asyncio.gather(
        *[start_server(platform=platform, server_key=server_key) for platform in servers]
    )


def run(
    server_key: str,
    server: Optional[str] = None,
    loop: Optional[asyncio.base_events.BaseEventLoop] = None,
) -> None:
    """
    Run server(s)

    Args:
        server_key: string path to key to use for ssh server(s)
        server: name of server to run, i.e. "cisco_iosxe", if None, run all servers
        loop: event loop to use, if none provided get current loop

    Returns:
        None

    Raises:
        N/A

    """
    if loop is None:
        loop = asyncio.get_event_loop()
    if server is None:
        servers = SERVERS
    else:
        servers = [server]
    loop.run_until_complete(main(server_key=server_key, servers=servers))
    loop.run_forever()



"""scrapli.driver.core.arista_eos.base_driver"""
import re
from typing import Dict

from scrapli.driver.base_network_driver import PrivilegeLevel

PRIVS = {
    "exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-@()/: ]{1,63}>\s?$",
            name="exec",
            previous_priv="",
            deescalate="",
            escalate="",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "privilege_exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-@/: ]{1,63}#\s?$",
            name="privilege_exec",
            previous_priv="exec",
            deescalate="disable",
            escalate="enable",
            escalate_auth=True,
            escalate_prompt=r"^[pP]assword:\s?$",
        )
    ),
    "configuration": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-@/: ]{1,63}\(config(?!\-s\-)[a-z0-9_.\-@/:]{0,32}\)#\s?$",
            name="configuration",
            previous_priv="privilege_exec",
            deescalate="end",
            escalate="configure terminal",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
}


class EOSDriverBase:
    # EOSDriverBase Mixin values set in init of sync/async NetworkDriver classes
    privilege_levels: Dict[str, PrivilegeLevel]

    def _create_configuration_session(self, session_name: str) -> None:
        """
        Handle configuration session creation tasks for consistency between sync/async versions

        Args:
            session_name: name of session to register

        Returns:
            N/A:  # noqa: DAR202

        Raises:
            ValueError: if a session of given name already exists

        """
        if session_name in self.privilege_levels.keys():
            msg = (
                f"session name `{session_name}` already registered as a privilege level, chose a "
                "unique session name"
            )
            raise ValueError(msg)
        session_prompt = re.escape(session_name[:6])
        pattern = (
            rf"^[a-z0-9.\-@/:]{{1,32}}\(config\-s\-{session_prompt}[a-z0-9_.\-@/:]{{0,32}}\)#\s?$"
        )
        name = session_name
        config_session = PrivilegeLevel(
            pattern=pattern,
            name=name,
            previous_priv="privilege_exec",
            deescalate="end",
            escalate=f"configure session {session_name}",
            escalate_auth=False,
            escalate_prompt="",
        )
        self.privilege_levels[name] = config_session



"""mock_devices.cisco_nxos"""
from typing import Any

from .base import BaseServer, BaseSSHServerSession

SHOW_COMMAND = """Cisco Nexus Operating System (NX-OS) Software
TAC support: http://www.cisco.com/tac
Documents: http://www.cisco.com/en/US/products/ps9372/tsd_products_support_series_home.html
Copyright (c) 2002-2019, Cisco Systems, Inc. All rights reserved.
The copyrights to certain works contained herein are owned by
other third parties and are used and distributed under license.
Some parts of this software are covered under the GNU Public
License. A copy of the license is available at
http://www.gnu.org/licenses/gpl.html.

Nexus 9000v is a demo version of the Nexus Operating System

Software
  BIOS: version
 NXOS: version 9.2(4)
  BIOS compile time:
  NXOS image file is: bootflash:///nxos.9.2.4.bin
  NXOS compile time:  8/20/2019 7:00:00 [08/20/2019 15:52:22]


Hardware
  cisco Nexus9000 9000v Chassis
   with 6096432 kB of memory.
  Processor Board ID 9GXMY611NJ4

  Device name: switch
  bootflash:    3509454 kB
Kernel uptime is 0 day(s), 1 hour(s), 4 minute(s), 47 second(s)

Last reset
  Reason: Unknown
  System version:
  Service:

plugin
  Core Plugin, Ethernet Plugin

Active Package(s):

"""


class NXOSSSHServerSession(BaseSSHServerSession):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Cisco NXOS mock ssh server

        Args:
            args: args to pass to base class
            kwargs: keyword args to pass to base class

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(*args, **kwargs)

        self.terminal_length = 34

        self.priv_levels = {
            "privilege_exec": "switch#",
            "config": "switch(config)#",
        }
        self.priv_level = "privilege_exec"
        self.invalid_input = "% Invalid command at '^' marker.\n"
        self.paging_prompt = "\n--More--"

        self.command_mapper = {
            self.handle_open: ["terminal length 0", "terminal width 511"],
            self.handle_close: ["exit"],
            self.handle_priv: ["enable", "configure terminal", "end"],
            self.handle_show: ["show version"],
            self.handle_interactive: [],
        }

        self.show_command_output = SHOW_COMMAND

    def handle_open(self, channel_input: str) -> None:
        """
        Handle "open" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level == "config":
            self.handle_invalid(channel_input=channel_input)
        elif channel_input == "terminal length 0":
            # in the future can actually set the value if we care?
            self.terminal_length = 0
            self._chan.write("")
            self._chan.write(self.priv_levels.get(self.priv_level))
        elif channel_input == "terminal width 511":
            self.terminal_width = 511
            self._chan.write("")
            self._chan.write(self.priv_levels.get(self.priv_level))
        else:
            self.handle_invalid(channel_input=channel_input)

    def handle_close(self, channel_input: str) -> None:
        """
        Handle "close" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _ = channel_input
        if self.priv_level == "config":
            # exit is a weird situation since it can be used to exit config mode or exit the device
            # so this lives here instead of handle_priv
            self._chan.write("")
            self.priv_level = "privilege_exec"
            self.repaint_prompt()
        else:
            self.eof_received()

    def _handle_priv_configure(self, channel_input: str) -> None:
        """
        Handle the "configure terminal" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level != "privilege_exec":
            self.handle_invalid(channel_input=channel_input)
        else:
            self._chan.write("")
            self.priv_level = "config"
            self.repaint_prompt()

    def _handle_priv_end(self, channel_input: str) -> None:
        """
        Handle the "end" command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.priv_level != "config":
            self.handle_invalid(channel_input=channel_input)
        else:
            self._chan.write("")
            self.priv_level = "privilege_exec"
            self.repaint_prompt()

    def handle_priv(self, channel_input: str) -> None:
        """
        Handle "privilege" commands sent to the ssh channel

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if channel_input == "configure terminal":
            self._handle_priv_configure(channel_input=channel_input)
        elif channel_input == "end":
            self._handle_priv_end(channel_input=channel_input)

    def handle_interactive(self, channel_input: str) -> None:
        """
        Handle the "interactive" (clear logging) command

        Args:
            channel_input: command sent to the ssh channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        # TODO
        pass

    def data_received(self, data: str, datatype: None) -> None:
        """
        Handle data received on ssh channel

        Args:
            data: string of data sent to channel
            datatype: dunno! in base class though :)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        # in the future we can cutoff the inputs if it is over X width if disable width has not yet
        # been ran -- not needed now but could be cool; if we just send a return, we should NOT
        # strip that!
        channel_input = data if data == "\n" else data.rstrip()

        if data == "\n":
            self.repaint_prompt()
            return

        for handler, channel_inputs in self.command_mapper.items():
            if channel_input in channel_inputs:
                # as soon as we find a match for the channel input ship it to the handler; handler
                # can figure out if command is valid for the given priv and such
                handler(channel_input=channel_input)
                return

        self.handle_invalid(channel_input=channel_input)


class NXOSServer(BaseServer):
    pass
